"""
Base pydantic models used by Abipy to implement the HTC version of the Flow connected to a MongoDB database.
"""
from __future__ import annotations

import json
import os
import inspect
import time
import zlib

#import numpy as np

from typing import Any, Type, TypeVar, Optional, Iterable
#from uuid import UUID
#from pprint import pprint
from pydantic import BaseModel, Field, PrivateAttr, root_validator  # , SecretStr
from pydantic.main import ModelMetaclass
from bson.objectid import ObjectId
from pymongo import MongoClient
from pymongo.database import Database
from pymongo.collection import Collection
from gridfs import GridFS
from monty.json import MontyEncoder, MontyDecoder, MSONable
from pymatgen.core import __version__ as pmg_version
from pymatgen.core.composition import Composition
from pymatgen.core.structure import Structure as pmg_Structure
from pymatgen.core.periodic_table import Element
from abipy.tools.serialization import pmg_serialize
from abipy.core.release import __version__ as abipy_version
from abipy.core.structure import Structure
from abipy.tools.iotools import ask_yes_no
from abipy.abio.inputs import AbinitInput
from abipy.electrons.ebands import ElectronBands
from abipy.flowtk.events import EventReport


class AbipyDecoder(MontyDecoder):
    """
    Extends MontyDecoder adding support for decoding classes besides instances..
    """

    def process_decoded(self, d):

        if isinstance(d, dict) and "@module" in d and "@qualname" in d:
            modname = d["@module"]
            qualname = d["@qualname"]
            mod = __import__(modname, None, None, [qualname], 0)
            return getattr(mod, qualname)

        return super().process_decoded(d)


def cls2dict(cls) -> dict:
    """Encodes python class using __qualname__"""
    return {"@qualname": cls.__qualname__, "@module": cls.__module__}


class AbipyEncoder(MontyEncoder):
    """
    Extends MontyEncoder adding support for encoding classes besides instances..
    """

    def default(self, o) -> dict:  # pylint: disable=E0202
        if inspect.isclass(o):
            return cls2dict(o)

        return super().default(o)


def monty_json_dumps(obj: Any, **kwargs) -> str:
    #print("in monty json dumps with type:", type(obj))
    return json.dumps(obj, cls=AbipyEncoder, **kwargs)


def monty_json_loads(string: str, **kwargs) -> Any:
    #print("in monty json loads with string", string)
    return json.loads(string, cls=AbipyDecoder, **kwargs)


def monty_load(obj: Any, **kwargs) -> dict:
    s = monty_json_dumps(obj)
    #print("in monty load:\n", s)
    d = monty_json_loads(s)
    #print(d)
    return d


def monty_trick(obj: Any) -> Any:
    #print(f"in monty trick with {type(obj)}")
    return json.loads(monty_json_dumps(obj))


M = TypeVar('M', bound='AbipyModel')


class AbipyModel(BaseModel, MSONable):
    """
    Base class providing tools to serialize Abipy/Pymatgen objects
    implementing the MSONable protocol.
    Most of the AbiPy models should inherith from this class so that we
    can take advantage of pydantic Config to encode/decode objects that are not pydantci models.
    """

    class Config:

        # This is needed to be able to use MSONable AbiPy objects
        # such as ElectronBands that do not inherit from BaseModel
        arbitrary_types_allowed = True

        # Here we register specialized encoders to convert MSONable objects to JSON strings.
        # Subclasses will inherit these json_encoders.
        json_encoders = {
            ModelMetaclass: lambda cls: cls2dict(cls),
            #ObjectId: lambda oid: str(oid),
            ObjectId: lambda oid: monty_trick(oid),
            #np.ndarray: lambda arr: arr.tolist(),
            #UUID: lambda uuid: str(uuid),
            Element: lambda o: monty_trick(o),
            Composition: lambda o: monty_trick(o),
            pmg_Structure: lambda o: monty_trick(o),
            Structure: lambda o: monty_trick(o),
            AbinitInput: lambda o: monty_trick(o),
            ElectronBands: lambda o: monty_trick(o),
            EventReport: lambda o: monty_trick(o),
        }

        # whether to populate models with the value property of enums, rather than the raw enum.
        #use_enum_values = True


        # make pydantic raise a ValidationError if Model get unexpected arguments.
        extra = "forbid"

        #json_loads = monty_json_loads
        #json_dumps = monty_json_dumps

    @classmethod
    def from_dict(cls: Type[M], d: dict) -> M:
        """
        Reconstruct the model from a dictionary taking into account the MSONAble protocol
        """
        #print("d1 in from dict:")
        #pprint(d)
        d = monty_load(d)
        if isinstance(d, cls):
            return d
        #print("d2 after monty_load:")
        #pprint(d)
        return cls(**d)

    @pmg_serialize
    def as_dict(self) -> dict:
        """
        Convert the model to dictionary taking into account the MSONable protocol.
        """
        #return monty_trick(self)
        d = self.dict()
        #print("d in as dict:")
        #pprint(d)
        return d

    #@staticmethod
    #def serialize(self, data):

    def to_json(self, **kwargs) -> str:
        """
        Returns a json string representation of the MSONable object.
        """
        # NB: we pass kwargs unlike in monty to_json
        return json.dumps(self.as_dict(), cls=AbipyEncoder, **kwargs)

    @classmethod
    def from_json_file(cls: Type[M], filepath: str) -> M:
        """
        Helper function to reconstruct the model from a json file.
        """
        with open(filepath, "rt") as fp:
            return cls.from_json(fp.read())

    @classmethod
    def from_json(cls: Type[M], json_string: str) -> M:
        """
        Helper function to reconstruct the model from a json string.
        """
        new = monty_json_loads(json_string)
        if isinstance(new, cls): return new
        return cls(**new)

    def json_write(self, filepath: str, **kwargs) -> None:
        """
        Helper function to write a json string with the model to file.
        """
        with open(filepath, "wt") as fp:
            #fp.write(self.json(**kwargs))
            fp.write(self.to_json(**kwargs))

    #@abstractmethod
    #def get_panel_view(self, mng_connector: MongoConnector):
    #    """Return panel object with a view of the model"""

    def yaml_dump(self) -> str:
        import ruamel.yaml as yaml
        json_string = self.to_json()
        return yaml.safe_dump(yaml.safe_load(json_string), default_flow_style=False)


class GfsDesc(AbipyModel):
    """
    Base model with metadata needed to retriveve an entry from GridFs
    This submodel lives in a standard MongoDB collection and stores all the information
    Users do not need to instanciate this object directly as they should use the
    API provided by MongoConnector.
    """

    oid: ObjectId = Field(..., description="ID of the file in the GridFS collection.")

    collection_name: str = Field(..., description="Name of the GridFS collection in which the file is stored.")

    zlib_level: int = Field(..., description="zlib compression level")


class GfsFileDesc(GfsDesc):
    """
    A Descriptor for files in GridFs.
    """

    filepath: str = Field(..., description="Absolute path to the file (depends on the host where the file has been produced")


class MongoConnector(AbipyModel):
    """
    Stores the parameters used to connect to the MongoDB server and the name of
    the collection used to perform CRUD operations.
    This object is usually instanciated from the AbiPy configuration file that defines the host,
    port, dbname, username and password.
    Note, however, that we don't store the name of the collection in the configuration file.
    so the user has to specify it explictly when calling ``from_abipy_config``.
    """

    host: str = Field(..., description="Host address e.g. 'localhost'")

    port: int = Field(27017, description="MongoDB server port")

    dbname: str = Field("abipy", description="MongoDB database")

    collection_name: str = Field(..., description="Default MongoDB collection")
    #gfs_collection_name: str = Field(..., description="Default GridFs collection")

    username: str = Field(None, description="User name for MongDB authentication. Implies password.")

    password: str = Field(None, description="Password for MongDB authentication")

    # Private attributes
    _client: MongoClient = PrivateAttr(None)

    @classmethod
    def for_localhost(cls, collection_name: str, port: int = 27017, **kwargs) -> MongoConnector:
        """
        Build an instance assuming a MongoDB server running on localhost listening on the default port
        """
        d = dict(host="localhost", port=port, collection_name=collection_name)
        d.update(kwargs)
        return cls(**d)

    @classmethod
    def from_abipy_config(cls, collection_name: str, **kwargs) -> MongoConnector:
        """
        Build an instance from the configuration options stored in the AbiPy config.yml file.
        """
        from abipy.core.config import get_config
        config = get_config()
        d = dict(host=config.mongo_host, port=config.mongo_port,
                 username=config.mongo_username, password=config.mongo_password,
                 dbname=config.mongo_dbname,
                 collection_name=collection_name,
                 )
        d.update(kwargs)

        return cls(**d)

    def new_with_collection_name(self, collection_name: str) -> MongoConnector:
        """
        Create a new MongoConnector with a different collection_name
        """
        d = self.dict()
        d["collection_name"] = collection_name
        return self.__class__(**d)

    @root_validator
    def check_username_and_password(cls, values):
        """Both Username and password are needed"""
        username, password = values.get("username"), values.get("password")
        if username is not None and password is None:
            raise ValueError('password must be provided when username is not None')
        return values

    def __str__(self) -> str:
        return self._repr_markdown_()

    def _repr_markdown_(self) -> str:
        """Markdown representation."""
        return f"""

## MongoConnector

- Host: {self.host}
- Port: {self.port}
- Database: {self.dbname}
- Collection: {self.collection_name}
- User: {self.username}
"""

    def get_client(self) -> MongoClient:
        """
        Establish a connection with the MongoDB server. Return MongoClient
        """
        from pymongo.errors import ConnectionFailure

        def client_is_connected(client: MongoClient) -> bool:
            try:
                # The ping command is cheap and does not require auth.
                client.admin.command("ping")
                return True
            except ConnectionFailure:
                print("Server not available. Trying to reconnect...")
                return False

        if self._client is not None and client_is_connected(self._client):
            return self._client

        # Reconnect and cache new client.
        if self.username and self.password:
            # Authenticate if needed
            self._client = MongoClient(host=self.host,
                                       port=self.port,
                                       username=self.username,
                                       password=self.password,
                                       authSource=self.dbname,
                                       #authMechanism='SCRAM-SHA-256'
                                       )

        else:
            self._client = MongoClient(host=self.host, port=self.port)

        return self._client

    def get_db(self) -> Database:
        """Return MongoDB database."""
        client = self.get_client()
        return client[self.dbname]

    def get_collection(self, collection_name: Optional[str] = None) -> Collection:
        """
        Returns MongoDB collection from its name.
        Use default collection if ``collection_name`` is None.
        """
        db = self.get_db()
        return db[collection_name or self.collection_name]

    def get_gridfs_and_name(self, collection_name: Optional[str] = None, ret_collname=False, **kwargs) -> tuple[GridFS, str]:
        """
        Returns GridFS collection. Use automatically generated name.
        """
        db = self.get_db()
        coll_name = f"{self.collection_name}_gridfs" if collection_name is None else collection_name
        fs = GridFS(db, collection=coll_name, **kwargs)
        return fs, coll_name

    def list_collection_names(self) -> list[str]:
        """
        "Return list of strings with all collection names in the database.
        """
        db = self.get_db()
        _filter = {"name": {"$regex": r"^(?!system\.)"}}
        return db.list_collection_names(filter=_filter)

    def list_collections(self) -> Iterable[Collection]:
        """
        Return iterable with collections available in the MongoDB database.
        """
        db = self.get_db()
        return db.list_collections()

    def drop_collection(self, ask_for_confirmation: bool = True) -> bool:
        """
        Drop the collection and the GridFs associated to it.
        If ask_for_confirmation is True (default), the user is prompted for confirmation.
        """
        if ask_for_confirmation:
            confirm = ask_yes_no(
                    f"Do you really want to drop collection: `{self.collection_name}` and the associated GridFs? [y/n]",
                    default=None)
            if not confirm:
                print("Command aborted by user. Your collection is safe.")
                return False

        db = self.get_db()
        collection = self.get_collection()
        #if collection.count_documents({}) == 0:
        #    print(f"Empty collection: `{collection.name}`")
        #    return True

        print(f"Dropping collection: `{collection.name}`")
        fs, fs_collname = self.get_gridfs_and_name()

        db.drop_collection(f"{fs_collname}.chunks")
        db.drop_collection(f"{fs_collname}.files")
        collection.drop()
        print("The MongoDB collection and the associated GridFs collection have been removed")

        return True

    def gfs_put_mson_obj(self, obj, zlib_level: int = zlib.Z_DEFAULT_COMPRESSION) -> GfsDesc:
        """
        Insert a MSONable object in GridFS
        """
        s = monty_json_dumps(obj)
        encoding = "utf-8"
        if zlib_level != zlib.Z_NO_COMPRESSION:
            s = zlib.compress(s.encode(encoding), level=zlib_level)

        metadata = {
            "@module": obj.__class__.__module__,
            "@class": obj.__class__.__name__,
            "zlib_level": zlib_level,
            "encoding": encoding,
        }

        #print(f"with zlib_level {zlib_level} len(s): {len(s)}")
        fs, fs_collname = self.get_gridfs_and_name()
        oid = fs.put(s, **metadata)

        return GfsDesc(oid=oid, collection_name=fs_collname, zlib_level=zlib_level)

    def gfs_get_mson_obj(self, gfsd: GfsDesc) -> Any:
        """
        Get a MSONable object from GridFS.
        """
        if gfsd.oid is None:
            raise RuntimeError("gfsd.oid is None, this means that the object has not been inserted in GridFS")

        fs, fs_collname = self.get_gridfs_and_name()
        # FIXME
        assert fs_collname == gfsd.collection_name

        grid_out = fs.get(gfsd.oid)
        s = grid_out.read()
        if gfsd.zlib_level != zlib.Z_NO_COMPRESSION:
            s = zlib.decompress(s)

        return monty_json_loads(s)

    def gfs_put_filepath(self, filepath: str, zlib_level: int = zlib.Z_DEFAULT_COMPRESSION) -> GfsFileDesc:
        """
        Insert a file in the default GridFs collection.
        Build and return a GfsFileDesc instance with metadata and the id.
        """
        if not filepath:
            raise ValueError("Expecting nonnull filepath but got `{filepath}`")

        fs, fs_collname = self.get_gridfs_and_name()

        metadata = dict(
            filename=os.path.basename(filepath),
            filepath=filepath,
            zlib_level=zlib_level,
        )

        if zlib_level == zlib.Z_NO_COMPRESSION:
            # This requires less memory as GridFs will get data from file.
            with open(filepath, "rb") as fh:
               oid = fs.put(fh, **metadata)
        else:
            # Here we need to read all the content in order to compress.
            with open(filepath, "rb") as fh:
                s = zlib.compress(fh.read(), level=zlib_level)
                oid = fs.put(s, **metadata)

        return GfsFileDesc(filepath=filepath, oid=oid, collection_name=fs_collname, zlib_level=zlib_level)

    def mktmp_filepath(self, gfsd: GfsFileDesc) -> str:
        """
        Fetch the file from GridFs, write it to a temporay file and return the filepath.
        """
        if gfsd.oid is None:
            raise RuntimeError("gfsd.oid is None, this means that the file has not been inserted in GridFS")

        fs, fs_collname = self.get_gridfs_and_name()
        # FIXME
        assert fs_collname == gfsd.collection_name

        grid_out = fs.get(gfsd.oid)

        from tempfile import mkstemp #, TemporaryFile, NamedTemporaryFile
        _, filepath = mkstemp(suffix=grid_out.filename)
        with open(filepath, "wb") as fh:
            s = grid_out.read()
            if grid_out.zlib_level != zlib.Z_NO_COMPRESSION:
                s = zlib.decompress(s)
            fh.write(s)

        return filepath

    def abiopen_gfsd(self, gfsd: GfsFileDesc):
        """
        Get binary data from the GridFS collection, write buffer to temporary file
        and use ``abiopen`` to open the file.
        """
        from abipy.abilab import abiopen
        filepath = self.mktmp_filepath(gfsd)
        return abiopen(filepath)

    def open_mongoflow_gui(self, flow_model_cls=None, **serve_kwargs):
        """
        Start panel GUI that allows the user to inspect the results stored in the collection.
        """
        if flow_model_cls is None:
            collection = self.get_collection()
            from .flow_models import FlowModel
            flow_model_cls = FlowModel.get_subclass_from_collection(collection)

        from abipy.panels.core import abipanel
        from abipy.panels.mongo_gui import MongoGui
        abipanel()
        app = MongoGui(self, flow_model_cls).get_app()
        import panel as pn
        pn.serve(app, **serve_kwargs)

    #def get_panel_view(self, mng_connector: MongoConnector):
    #    import panel as pn
    #    return pn.pane.HTML(self._repr_markdown_())

    ############################
    # FlowModel specific methods
    ############################

    def init_flow_model_collection(self, flow_model_cls, protocol=None) -> None:

        if not hasattr(flow_model_cls, "_magic_key"):
            raise TypeError(f"Expecting FlowModel subclass with `_magic_key, got {flow_model_cls.__name__}")

        new_dict = cls2dict(flow_model_cls)
        if protocol:
            new_dict.update({"protocol": protocol.dict()})

        collection = self.get_collection()

        cnt, oid = 0, None
        for doc in collection.find({flow_model_cls._magic_key: {"$exists": True}}):
            doc.pop("_id")
            old_dict = doc.get(flow_model_cls._magic_key)
            cnt += 1
            #if new_dict != old_dict:
            #    raise TypeError(f"Cannot register new FlowModel:\n\n\t{new_dict}\n\n"
            #                    f"as collection `{collection.name}` is already associated to FlowModel:\n\n"
            #                    f"\t{old_dict}")

        if cnt != 0:
            raise ValueError("Collection is already init")

        # Create new document with FlowModel class if not already present.
        collection.insert_one({flow_model_cls._magic_key: new_dict})

    #def get_flow_model_cls_and_protocol(self):
    #    collection = self.get_collection()

    #    key = flow_model_cls._magic_key

    #    cnt, oid = 0, None
    #    for doc in collection.find({key: {"$exists": True}}):
    #        doc.pop("_id")
             #data = AbipyDecoder().process_decoded(doc)
    #        old_dict = doc.get(key)
    #        cnt += 1

    #    return flow_model_cls, protocol

    def insert_flow_models(self, models: list[TopLevelModel], verbose: int = 0) -> list[ObjectId]:
        """
        Insert list of FlowModels in collection.
        Return list of objectid
        """
        flow_model_cls = models[0].__class__

        # Sanity check. We need a list of FlowModel objects of the same type.
        if any(m.__class__ is not flow_model_cls for m in models):
            raise ValueError(f"All models must belong to the same class {models[0].__class__}")

        if not hasattr(flow_model_cls, "_magic_key"):
            raise TypeError(f"Expecting FlowModel subclass with `_magic_key, got {flow_model_cls.__name__}")

        new_dict = cls2dict(flow_model_cls)

        collection = self.get_collection()
        cnt, oid = 0, None
        for doc in collection.find({flow_model_cls._magic_key: {"$exists": True}}):
            doc.pop("_id")
            old_dict = doc.get(flow_model_cls._magic_key)
            cnt += 1
            if new_dict != old_dict:
                raise TypeError(f"Cannot register new FlowModel:\n\n\t{new_dict}\n\n"
                                f"as collection `{collection.name}` is already associated to FlowModel:\n\n"
                                f"\t{old_dict}")

        if cnt == 0:
            # Create new document with FlowModel class if not already present.
            collection.insert_one({flow_model_cls._magic_key: new_dict})

        return mng_insert_models(models, collection)

    def find_err_oid_flowmodels(self) -> Iterable[tuple]:
        collection = self.get_collection()

        from .flow_models import FlowModel
        flow_model_cls = FlowModel.get_subclass_from_collection(collection)

        query = {"flow_data.exec_status": "Errored"}
        for doc in collection.find(query):
            data = AbipyDecoder().process_decoded(doc)
            oid = data.pop("_id")
            yield (oid, flow_model_cls(**data))


class MockedMongoConnector(MongoConnector):
    """
    Mock a MongoConnector using mongomock. Mainly used for unit tests.
    """

    def get_client(self) -> MongoClient:
        # Cache the client also in the Mock so that we use the same mocked instance.
        if self._client is not None:
            return self._client

        import mongomock
        self._client = mongomock.MongoClient()
        return self._client

    #def get_collection(self, collection_name: Optional[str] = None) -> Collection:
    #    return mongomock.MongoClient().db.collection

    def get_gridfs_and_name(self, collection_name: Optional[str] = None, **kwargs) -> tuple[GridFS, str]:
        from mongomock.gridfs import enable_gridfs_integration
        enable_gridfs_integration()
        return super().get_gridfs_and_name(collection_name=collection_name, **kwargs)


class TopLevelModel(AbipyModel):
    """
    Provides tiny wrappers around the MongoDB API.
    """

    abipy_version: str = Field(
        abipy_version, description="The version of abipy this document was built with"
    )

    pymatgen_version: str = Field(
        pmg_version, description="The version of pymatgen this document was built with"
    )

    @classmethod
    def from_oid(cls, oid: ObjectId, collection: Collection):
        """
        Return the model from the ObjectId oid and the MongoDB collection.
        """
        data = collection.find_one({'_id': oid})
        if not data:
            raise ValueError(f"Cannot find {cls.__name__} model in collection: {collection}")

        data.pop("_id")
        data = AbipyDecoder().process_decoded(data)

        #if "scf_data" in data: print(data["scf_data"])

        return cls(**data)

    @classmethod
    def mng_find(cls, query: dict, collection: Collection, **kwargs) -> QueryResults:
        """
        Find the models in the collection matching the mongodb query.
        """
        cursor = collection.find(query, **kwargs)

        oids, models = [], []
        decoder = AbipyDecoder()
        for data in cursor:
            oids.append(data.pop("_id"))
            data = decoder.process_decoded(data)
            models.append(cls(**data))

        return QueryResults(oids, models, query)

    def mng_insert(self, collection: Collection) -> ObjectId:
        """
        Insert the model in the collection. Return ObjectId.
        """
        # self.json build a JSON string: pydantic handles the serialization of the pydantic models
        # while arbitrary types (e.g. Structure) are encoded using json_encoders.
        # json loads allows us to get the final dictionary to be inserted in the MongoDB collection.
        doc = json.loads(self.json())
        return collection.insert_one(doc).inserted_id

    def mng_full_update_oid(self, oid: ObjectId, collection: Collection) -> None:
        """
        Perform a full update of the document given the ObjectId `oid` in `collection`.
        """
        # TODO: Implement Atomic transaction
        old_doc = collection.find_one({'_id': oid})
        if not old_doc:
            raise ValueError(f"Cannot find document with ObjectId: {oid}")

        new_doc = json.loads(self.json())
        old_doc.update(new_doc)

        collection.replace_one({"_id": oid}, new_doc, upsert=False)

    #def backup_oid_collection_name(self, oid: ObjectId, collection_name: str) -> str:
    #    """
    #    Write a backup file in the abipy HOME directory. Return path to the file.

    #    Useful if the MongoDB server goes down and the AbipyWorker needs
    #    to save the results somewhere on the filesystem.
    #    """
    #    bkp_dir = os.path.join(os.path.expanduser("~"), ".abinit", "abipy", "bkp_models")
    #    if not os.path.isdir(bkp_dir): os.mkdir(bkp_dir)
    #    filename = f"{collection_name}_{str(oid)}"
    #    filepath = os.path.join(bkp_dir, filename)
    #    with open(filepath, "wt") as fp:
    #        fp.write(self.json())
    #        return filepath

#def read_backup_oid_collection_name(cls, oid, collection_name):
#    bkp_dir = os.path.join(os.path.expanduser("~"), ".abinit", "abipy", "bkk_models")
#    filepath = os.path.join(bkp_dir, f"{collection_name}_{str(oid)}")
#    if not os.path.exists(filepath):
#    return cls.from_json_file(filepath)


def mng_insert_models(models: list[AbipyModel], collection: Collection, verbose: int = 0) -> list[ObjectId]:
    """
    Insert list of models in a collection. If verbose > 0, print elasped time.
    Return: list of ObjectId
    """
    start = time.time()
    docs = [json.loads(model.json()) for model in models]
    oids = collection.insert_many(docs).inserted_ids
    if verbose:
        print(f"Inserted {len(oids)} MongoDB documents in collection: {collection.full_name} in "
              f"{time.time() - start: .2f} [s]")
    return oids


class QueryResults:
    """
    Stores the results of a MongoDB query in the form of modelds instead of plain dictionaries.
    """

    def __init__(self, oids: list[ObjectId], models: list[AbipyModel], query: dict) -> None:
        """
        Args:
            oids:
            models:
            query:
        """
        self.oids = oids
        self.models = models
        self.query = query

    @classmethod
    def empty_from_query(cls, query: dict) -> QueryResults:
        """
        Return an empty QueryResults produced by query.
        """
        return cls(oids=[], models=[], query=query)

    def __len__(self) -> int:
        return len(self.oids)

    def __bool__(self) -> bool:
        return bool(self.oids)



#!/usr/bin/env python
from __future__ import annotations

from typing import List

from pydantic import BaseModel


def _find_submodels(self, class_type):
    """
    Warning: does not handle containers (list or dict)
    with models e.g. [instance_of_class_type,  ...]
    """
    desc_list = []
    for k, v in self:
        #print(type(k), type(v))
        print("k:", k, "v:", v)
        if isinstance(v, class_type):
            desc_list.append(v)

        elif isinstance(v, BaseModel):
            sub_list = _find_submodels(v, class_type)
            desc_list.extend(sub_list)

        elif isinstance(v, (list, tuple)):
            for o in v:
                if isinstance(o, class_type):
                    desc_list.append(o)
                elif isinstance(o, BaseModel):
                    sub_list = _find_submodels(o, class_type)
                    desc_list.extend(sub_list)

    return desc_list


def find_submodels(self, class_type):
    desc_list = _find_submodels(self, class_type)
    # Remove duplicated objects if any
    d = {id(desc): desc for desc in desc_list}
    return list(d.values())


class NestedModel(BaseModel):
    foo: int


class MyModel(BaseModel):
    bar: str = "hello"
    one: NestedModel
    two: NestedModel

class TopLevelModel(BaseModel):
    my_model1: MyModel
    info: str = "This is the top level model"
    my_model2: MyModel


my_model = MyModel(one=NestedModel(foo=1), two=NestedModel(foo=2))

mlist = find_submodels(my_model, NestedModel)
#print(mlist)
assert len(mlist) == 2 and all(isinstance(mod, NestedModel) for mod in mlist)
for mod in mlist:
    mod.foo *= 100

assert my_model.one.foo == 100
assert my_model.two.foo == 200

# Note that here we have two references to same object.
top_model = TopLevelModel(my_model1=my_model, my_model2=my_model)
mlist = find_submodels(top_model, NestedModel)
print(mlist)
assert len(mlist) == 2 and all(isinstance(mod, NestedModel) for mod in mlist)
for mod in mlist:
    mod.foo = mod.foo / 100

assert top_model.my_model1.one.foo == 1
assert top_model.my_model1.two.foo == 2


class ModelWithList(BaseModel):
    info: str = "This is the top level model"
    mlist: List[MyModel]
    items: List[int]


model1 = MyModel(one=NestedModel(foo=1), two=NestedModel(foo=2))
model2 = MyModel(one=NestedModel(foo=20), two=NestedModel(foo=40))

model_with_list = ModelWithList(mlist=[model1, model2], items=[1,2,3])

mlist = find_submodels(model_with_list, NestedModel)
print(mlist)
assert len(mlist) == 4 and all(isinstance(mod, NestedModel) for mod in mlist)
#for mod in mlist:
#    mod.foo = mod.foo / 100



"""Panels for interacting with output files in text format."""
from __future__ import annotations

import panel as pn
import panel.widgets as pnw
import bokeh.models.widgets as bkw

from abipy.panels.core import AbipyParameterized, Loading, mpl, ply, dfc


class AbinitOutputFilePanel(AbipyParameterized):
    """
    Panel with widgets to interact with the main Abinit output file.
    """
    def __init__(self, outfile, **params):
        super().__init__(**params)
        self.outfile = outfile

    def get_cycles_view(self, what: str) -> pn.GridBox:
        """
        Return GridBox with matplotlib plots for the GS/DFPT SCF cycles
        or None if no cycle is found.
        """
        if what == "GS_SCF":
            cycles = self.outfile.get_all_gs_scf_cycles()
        elif what == "DFPT_D2DE":
            cycles = self.outfile.get_all_d2de_scf_cycles()
        else:
            raise ValueError(f"Invalid value for what: `{what}`")

        if not cycles: return None

        num_plots, nrows, ncols = len(cycles), 1, 1
        if num_plots > 1:
            ncols = 2
            nrows = (num_plots // ncols) + (num_plots % ncols)

        box = pn.GridBox(nrows=nrows, ncols=ncols, sizing_mode='stretch_width')
        #box = pn.Column(sizing_mode="stretch_width")

        for icycle, cycle in enumerate(cycles):
            #box.append(mpl(cycle.plot(title="%s cycle #%d" % (what, icycle), **self.mpl_kwargs)))
            f = ply(cycle.plotly(title="%s cycle #%d" % (what, icycle + 1), show=False))
            box.append(f)

        return box

    def get_outfile_view(self) -> pn.Column:
        col = pn.Column(sizing_mode="stretch_width")
        ca = col.append; cext = col.extend

        filepath = self.outfile.filepath
        with open(filepath, "rt") as fh:
            text = fh.read()

        ace = pnw.Ace(value=text, language='text', readonly=True,
                      sizing_mode='stretch_width', height=1200)
                      #sizing_mode='stretch_width', width=900)
        cext([f"## Output <small>{filepath}</small>", ace, pn.layout.Divider()])

        return col

    def get_panel(self, as_dict=False, **kwargs):
        """Return tabs with widgets to interact with the Abinit output file."""
        d = {}

        d["Summary"] = self.get_summary_view_for_abiobj(self.outfile)
        #d["Output"] = self.get_outfile_view()

        df = self.outfile.get_dims_spginfo_dataframe()
        df.index.name = "Dataset"
        d["Dims"] = dfc(df, transpose=True)

        # Add tabs with plots for the GS/DFPT SCF cycles.
        for what in ("GS_SCF", "DFPT_D2DE"):
            box = self.get_cycles_view(what)
            if box is not None:
                d[f"{what} cycles"] = box

        #timer = self.get_timer()
        #timer.plot_all(**self.mpl_kwargs)

        if as_dict: return d

        return self.get_template_from_tabs(d, template=kwargs.get("template", None))


class AbinitOutputFilePanelWithFileInput(AbipyParameterized):

    info_str = """
This application allows users to analyze the Abinit main output file
"""

    def __init__(self, **params):
        super().__init__(**params)

        help_md = pn.pane.Markdown(f"""
## Description

{self.info_str}
""")

        self.main_area = pn.Column(help_md, sizing_mode="stretch_width")
        self.abifile = None

        self.file_input = pnw.FileInput(height=60, css_classes=["pnx-file-upload-area"])
        self.file_input.param.watch(self.on_file_input, "value")

    def on_file_input(self, event) -> None:

        with Loading(self.main_area):
            new_abifile = self.get_abifile_from_file_input(self.file_input)

            if self.abifile is not None:
                self.abifile.remove()

            self.abifile = new_abifile
            self.main_area.objects = [self.abifile.get_panel()]

    def get_panel(self, **kwargs):

        col = pn.Column("## Upload (or drag & drop) an *.abo* file (main ABINIT output file):",
                        self.get_fileinput_section(self.file_input),
                        sizing_mode="stretch_width")

        main = pn.Column(col, self.main_area, sizing_mode="stretch_width")

        cls, cls_kwds = self.get_abinit_template_cls_kwds()

        return cls(main=main, title="Abo Analyzer", **cls_kwds)




import numpy as np

#from abipy.electrons.denpot import DfptPotNcFile
from abipy.tools.numtools import transpose_last3dims
from abipy.iotools import xsf


from abipy.electrons.denpot import _NcFileWithField


class DfptPotNcFile(_NcFileWithField):
    """
    POT file produced by the DFPT part containing the
    first order derivative of the KS potential in real space.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: DfptPotNcFile
    """
    field_name = "first_order_potential"


class DfptQPotsAnalyzer(object):

    def __init__(self, potname_gamma, potnames_qdir):

        self.potfiles_qdir, qpts, pertcases = [], [], []
        for i, path in enumerate(potnames_qdir):
            ncfile = DfptPotNcFile.from_file(path)
            hdr = ncfile.hdr
            #print(ncfile)
            print("qptn:", hdr.qptn, "pertcase:", hdr.pertcase)
            self.potfiles_qdir.append(ncfile)
            pertcases.append(hdr.pertcase)
            qpts.append(hdr.qptn)

            if i == 0:
                nspden = hdr.nspden
                ngfft = ncfile.reader.read_ngfft3()
                print("ngggt", ngfft)
                structure = ncfile.reader.read_structure()
            #print(ncfile.vks1)

        # Consistency check.
        #for

        # Rearrange qpoints if they are not already ordered.

        # Build [nq, ] array with the potentials
        # number_of_components, number_of_grid_points_vector3, number_of_grid_points_vector2,
        # number_of_grid_points_vector1, real_or_complex_first_order_potential

        for i, ncfile in enumerate(self.potfiles_qdir):
            rc = ncfile.reader.read_dimvalue("real_or_complex_first_order_potential")
            var = ncfile.reader.read_variable("first_order_potential")

            print("var.shape:", var.shape)
            if rc == 1:
                arr = var[:] + 0j
            elif rc == 2:
                arr = var[..., 0] + 1j * var[..., 1]
            else:
                raise ValueError("Invalid real_or_complex_first_order_potential: %d" % rc)
            arr = np.reshape(arr, -1)

            if i == 0:
                s = arr.shape + (len(self.potfiles_qdir), )
                print("arr.shape:", arr.shape, "s:", s)
                values = np.empty(s, dtype=arr.dtype)

            values[:, i] = arr

        print(values.dtype, values.shape)

        # Compute q --> 0 limit by linear extrapolation.
        qlens = structure.reciprocal_lattice.reciprocal_lattice.norm(qpts)
        print(qlens)

        #from scipy.interpolate import InterpolatedUnivariateSpline
        #f = InterpolatedUnivariateSpline(x, y, w=None, bbox=[None, None], k=3, ext=0, check_finite=False)

        from scipy.interpolate import interp1d
        #f = interp1d(x, y, kind='linear', axis=-1, copy=True, bounds_error=None, fill_value=nan, assume_sorted=False)
        f = interp1d(qlens, values, kind='linear', axis=-1, fill_value='extrapolate')
                     #copy=True, bounds_error=None, fill_value=nan, assume_sorted=False)

        extrap_values = f(0)
        print(extrap_values.shape)

        ispden = 0
        for i in range(len(qlens) + 1):
            if i == len(qlens):
                datar = extrap_values
            else:
                datar = values[:, i]
            filename = "foo%d.xsf" % i
            datar.shape = (nspden,) + tuple(ngfft)
            with open(filename, mode="wt") as fh:
                xsf.xsf_write_structure(fh, structure)
                xsf.xsf_write_data(fh, structure, datar[ispden], add_replicas=True, cplx_mode="abs")

        self.potfile_gamma = DfptPotNcFile.from_file(potname_gamma)
        ncfile = self.potfile_gamma
        rc = ncfile.reader.read_dimvalue("real_or_complex_first_order_potential")
        var = ncfile.reader.read_variable("first_order_potential")
        datar = var[:] + 0j
        filename = "foo_gamma.xsf"
        with open(filename, mode="wt") as fh:
            xsf.xsf_write_structure(fh, structure)
            xsf.xsf_write_data(fh, structure, datar[ispden], add_replicas=True, cplx_mode="abs")


class DfptInterpAnalyzer(object):

    def __init__(self, dfpt_pot_filepath, v1qg_nolr_filepath, v1vq_withlr_filepath):
        ispden = 0
        with DfptPotNcFile.from_file(dfpt_pot_filepath) as ncfile:
            r = ncfile.reader
            structure = r.read_structure()
            hdr = ncfile.hdr
            pertcase_c = hdr.pertcase - 1
            idir = pertcase_c % 3
            ipert = (pertcase_c - idir) // 3
            print("q-point:", hdr.qptn, "Fortran pertcase:", hdr.pertcase, "C idir:", idir, "C ipert:", ipert)

            # On netcdf file, we have array of shape.
            # (number_of_components, number_of_grid_points_vector3, number_of_grid_points_vector2,
            #   number_of_grid_points_vector1, real_or_complex_first_order_potential)
            rc = r.read_dimvalue("real_or_complex_first_order_potential")
            var = r.read_variable("first_order_potential")
            #print("var.shape:", var.shape)
            if rc == 1:
                datar = var[ispden, ...] + 0j
            elif rc == 2:
                datar = var[ispden, ..., 0] + 1j * var[ispden, ..., 1]
            else:
                raise ValueError("Invalid real_or_complex_first_order_potential: %d" % rc)

        def xsf_write(filepath, datar):
            with open(filepath, mode="wt") as fh:
                xsf.xsf_write_structure(fh, structure)
                cplx_mode = "re"
                #cplx_mode = "im"
                #cplx_mode = "abs"
                xsf.xsf_write_data(fh, structure, datar, cplx_mode=cplx_mode)

        self.dfpt_pot = transpose_last3dims(datar)
        xsf_write("dfpt.xsf", self.dfpt_pot)

        from abipy.eph.v1qavg import V1qAvgFile
        d = {}

        for path, key in zip([v1qg_nolr_filepath, v1vq_withlr_filepath], ["nolr", "withlr"]):
            with V1qAvgFile(path) as ncfile:
                r = ncfile.reader
                ngfft = r.read_value("ngfft")
                d[key] = {}
                for var_name in ("v1r_interpolated", "v1r_lrmodel"):
                    # nctkarr_t("v1r_interpolated", "dp", "two, nfft, nspden, natom3")
                    var = r.read_variable(var_name)
                    datar = var[pertcase_c, ispden, :, 0] + 1j * var[pertcase_c, ispden, :, 1]
                    # Because we are reading Fortran data (z,y,x) and xsf_write_data expects (..., x, y, z)
                    datar = transpose_last3dims(datar)
                    datar = np.reshape(datar, ngfft)
                    d[key][var_name] = datar
                    assert np.all(datar.shape == self.dfpt_pot.shape)

                    path = key + "_" + var_name + ".xsf"
                    xsf_write(path, datar)

        other = d["withlr"]["v1r_interpolated"]
        datar = self.dfpt_pot - other
        print("DFT AVG Re", np.average(self.dfpt_pot.real), ", DFT AVG Im", np.average(self.dfpt_pot.imag))
        print("INTERP AVG Re", np.average(other.real), ", INTERP AVG Im", np.average(other.imag))
        print("diff wrt DFPT with LR model", np.sum(np.abs(datar)) / datar.size)
        xsf_write("dfpt_minus_withlr.xsf", datar)

        other = d["nolr"]["v1r_interpolated"]
        datar = self.dfpt_pot - other
        print("DFT AVG Re", np.average(self.dfpt_pot.real), ", DFT AVG Im", np.average(self.dfpt_pot.imag))
        print("NOLR INTERP AVG Re", np.average(other.real), ", NOLR INTERP AVG Im", np.average(other.imag))
        print("diff wrt DFPT without LR model", np.sum(np.abs(datar)) / datar.size)
        xsf_write("dfpt_minus_nolr.xsf", datar)

        datar = d["withlr"]["v1r_interpolated"] - d["nolr"]["v1r_interpolated"]
        xsf_write("interp_lrnolr_diff.xsf", datar)


if __name__ == "__main__":
    import sys
    #potname_gamma, potnames_qdir = sys.argv[1], sys.argv[2:]
    #ana = DfptQPotsAnalyzer(potname_gamma, potnames_qdir)

    dfpt_pot_filepath, v1qg_nolr_filepath, v1vq_withlr_filepath = sys.argv[1:]
    ana = DfptInterpAnalyzer(dfpt_pot_filepath, v1qg_nolr_filepath, v1vq_withlr_filepath)



#!/usr/bin/env python
r"""
Projected phonon DOS
====================

This example shows how to plot several phonon DOS on a grid.
We phonon DOS file produced by anaddb:

    trf2_5.out_PHDOS.nc: phonon DOS compute with anaddb.

See also tutorial/lesson_rf2.html
"""

#%%
# We start by defining a list with the paths to the PHDOS.nc files
# In this case, for simplicity, we use the same file but we must
# use different labels when adding them to the plotter with the add_phdos method.

from abipy import abilab
import abipy.data as abidata

phdos_paths = [abidata.ref_file("trf2_5.out_PHDOS.nc")]

plotter = abilab.PhononDosPlotter()
plotter.add_phdos("AlAs", phdos_paths[0])
plotter.add_phdos("Same AlAs", phdos_paths[0])

#%%
# To visualize the results on a grid with matplotlib, use:
plotter.gridplot(units="cm-1", tight_layout=True, title="Phonon DOS in cm-1 on grid")

#%%
# For the plotly version:
plotter.gridplotly(units="cm-1", title="Phonon DOS in cm-1 on grid")



"""
"""
from __future__ import annotations

#from typing import List  # Dict, Tuple
from abc import ABC, abstractmethod
from pprint import pformat
from pydantic import BaseModel, Field, validator
from bson.objectid import ObjectId
from pymatgen.io.abinit.pseudos import Pseudo, PseudoTable
from abipy.core.structure import Structure
from abipy.flowtk.psrepos import get_repo_from_name
from .base_models import AbipyModel, GfsFileDesc, GfsDesc, MongoConnector


#class PseudoSpecs(_PseudosProvider):
class _PseudosProvider(BaseModel, ABC):
    """
    Abstract base class enforcing the implementation of ``get_pseudos``
    """

    ps_type: str = Field(..., description="Pseudopotential type e.g. NC or PAW.")

    ps_generator: str = Field(..., description="Name of the Pseudopotential generator.")

    xc_name: str = Field(..., description="Name of the XC functional e.g. LDA, PBE, PBEsol.")

    relativity_type: str = Field(..., description="Scalar-relativistic (SR) or fully-relativistic (FR)")

    @abstractmethod
    def get_pseudos(self) -> PseudoTable:
        """Build and return the PseudoTable associated the model."""


#class PseudoSpecsWithRepo(_PseudosProvider):
class PseudoSpecs(_PseudosProvider):
    """
    This model stores the parameters needed to retrieve the PseudoPotential Table from a repository.
    The `repo_name` defines the full set of versioned pseudos while `table_name` defines the list of pseudos.
    included in the final Table.
    Note that the PseudoDojo/JTC project may provide multiple pseudos for the same atom.
    This is the reason why the final table with one pseudo per element is uniquely defined by
    the tuple (repo_name, table_name).

    .. code-block::

        specs = PseudoSpecs.from_repo_table_name("ONCVPSP-PBEsol-SR-PDv0.4", "standard")
    """

    repo_name: str = Field(..., description="Name of the pseudopotential repository.")

    table_name: str = Field(..., description="Name of the table e.g. normal, stringent, MD ...")

    project_name: str = Field(..., description="PD for PseudoDojo, JTH for JTH.")

    version: str = Field(..., description="Version of the repository.")

    @classmethod
    def from_repo_table_name(cls, repo_name: str, table_name: str) -> PseudoSpecs:
        """
        Build the object from a repository name and ``table_name``.
        """
        repo = get_repo_from_name(repo_name)
        data = dict(repo_name=repo_name, table_name=table_name)
        attr_list = ["ps_generator", "ps_type", "xc_name", "relativity_type", "project_name", "version"]
        data.update({aname: getattr(repo, aname) for aname in attr_list})

        return cls(**data)

    def __str__(self) -> str:
        return pformat(self.dict())
        #lines = []
        #return "\n".join(lines)

    @validator("relativity_type")
    def validate_relativity_type(cls, value):
        """
        Validate relativity_type
        """
        allowed_values = ("SR", "FR")
        if value not in allowed_values:
            raise ValueError(f"{value} not in {allowed_values}")

        return value

    def get_pseudos(self) -> PseudoTable:
        """
        Build and return the PseudoTable associated to the specs.
        """
        repo = get_repo_from_name(self.repo_name)
        if not repo.is_installed():
            print(f"Warning: Could not find {self.repo_name} installed in {repo.dirpath}.\n"
                  f"Will try to install it at runtime...")
            repo.install(verbose=1)

        # Note that the repo instance keeps an internal cache of tables.
        return repo.get_pseudos(table_name=self.table_name)


class MongoDbPseudos(_PseudosProvider):
    """
    A rather specialized provider that takes pseudopotentials by IDs from a MongoDB collection.
    This approach is used for testing/validanting "unofficial" pseudos.
    """

    oid_list: List[ObjectId] = Field(..., description="List of ids in collection")

    mng_connector: MongoConnector = Field(..., description="")

    #pseudo_model_cls =

    def get_pseudos(self) -> PseudoTable:
        """
        Build and return the PseudoTable associated to the specs.
        """
        collection = self.mng_connector.get_collection()
        #pseudo_model = pseudo_model_cls(**d)
        pseudos = []
        for oid in self.oid_list:
            d = collection.find_one({"_id": oid}, {"_id": 0})
            if d is None:
                raise ValueError(f"Cannot find pseudo with id {oid} in collection: {collection}")
            #pseudo_model = pseudo_model_cls(**d)
            p = pseudo_model.get_pseudo_from_db(self.mng_connector)
            pseudos.append(p)

        return PseudoTable(pseudos)



class BasePseudoModel(AbipyModel, ABC):
    """
    Abstract base class for a model associated to a single pseudopotential.
    Used for fetching the pseudopotential file from a MongoDB collection.
    """

    atomic_symbol: str = Field(..., description="")

    ps_type: str = Field(..., description="Pseudopotential type e.g. NC or PAW.")

    ps_generator: str = Field(..., description="Name of the Pseudopotential generator.")

    xc_name: str = Field(..., description="Name of the XC functional e.g. LDA, PBE, PBEsol.")

    relativity_type: str = Field(..., description="Scalar-relativistic (SR) or fully-relativistic (FR)")

    #psgen_input_string: str = Field(..., description="")

    #psgen_output_string: str = Field(..., description="")

    #output_gsfd: GfsFileDesc = Field(..., description="Pointer to the main output file in GridFs")

    @abstractmethod
    def get_pseudo_from_db(self, mng_connector: MongoConnector) -> Pseudo:
        """Build and return a Pseudo object from the model."""



class OncvpspModel(BasePseudoModel):
    """
    This model stores the pseudopotentials generated by oncvpsp with the corresponding
    input and output files.
    """

    psp8_gfsd: GfsFileDesc = Field(..., description="Pointer to the psp8 file in GridFs")
    psp8_md5: str = Field(..., description="md5 checksum for the psp8 file")

    upf_md5: str = Field(..., description="md5 checksum for the upf file")
    upf_gsfd: GfsFileDesc = Field(..., description="Pointer to the UPF file in GridFs")

    psml_md5: str = Field(..., description="md5 checksum for the psml file")
    psml_gsfd: GfsFileDesc = Field(..., description="Pointer to the PSML file in GridFs")

    #@validator("atomic_symbol")
    #def validate_atomic_symbol(cls, value):

    @classmethod
    def from_dirpath(cls, dirpath: str, mng_connector: MongoConnector) -> OncvpspModel:
        """
        Build a the model from a directory containing the output files produced by oncvpsp.
        """
        from abipy.flowtk.psrepos import md5_for_filepath
        dirpath = os.path.abspath(dirpath)
        paths = [os.path.join(dirpath, b) for b in os.listdir(dirpath)]

        def find_abspath_byext(ext: str):
            cnt, ret_apath = 0, None
            for p in paths:
                if p.endswith(ext):
                    ret_apath = p
                    cnt += 1

            if cnt == 1:
                return ret_apath
            if cnt > 1:
                raise RuntimeError(f"Found multiple files with extension {ext} in {paths}")

            raise RuntimeError(f"Cannot find file with extension {ext} in {paths}")

        data = {}
        apath = find_abspath_byext(".in")
        input_string = open(apath, "rt").read()
        #data["psgen_input_string"] = input_string
        apath = find_abspath_byext(".out")
        #data["psgen_output_string"] = open(apath, "rt").read()
        #data[output_gfsd"] = mng_connector.gfs_put_filepath(apath)

        data["atomic_symbol"] = "foo"

        for fmt in ["psp8", "upf", "psml"]:
            apath = find_abspath_byext("." + fmt)
            data[f"{fmt}_md5"] = md5_for_filepath(apath)
            #with open(apath, "rt") as fh:
            #    data[f"{fmt}_input_string"] = fh.read()

            data[f"{fmt}_gfsd"] = mng_connector.gfs_put_filepath(apath)

        return cls(**data)

    def get_pseudo_from_db(self, mng_connector: MongoConnector) -> Pseudo:
        """
        Build and return a Pseudo instance from the model.
        """
        filepath = mng_connector.mktmp_filepath(self.psp8_gfsd)
        return Pseudo.from_file(filepath)




#!/usr/bin/env python
r"""
Computation of Phonons, BECS and Eps_inf for MgO
==================================================

This example shows how to compute phonons with DFPT
Symmetries are taken into account: only q-points in the IBZ are generated.
The final results (out_DDB, out_DVDB) will be produced automatically at the end of the run
and saved in the ``outdata/`` directory of work[1].
"""

import sys
import os
import abipy.abilab as abilab
import abipy.data as abidata

from abipy import flowtk


def make_scf_input():
    """
    This function constructs the input file for the GS calculation:
    """

    # Initialize MgO structure from abinit variables.
    structure = abilab.Structure.from_abivars(
        acell=3 * [4.252718 * abilab.units.ang_to_bohr],
        rprim=[0.0000000000, 0.5000000000, 0.5000000000,
               0.5000000000, 0.0000000000, 0.5000000000,
               0.5000000000, 0.5000000000, 0.0000000000],
        natom=2,
        ntypat=2,
        typat=[1, 2],
        znucl=[12, 8],
        xred=[0.0000000000, 0.0000000000, 0.0000000000,
              0.5000000000, 0.5000000000, 0.5000000000]
    )

    # NC pseudos assumed in currect working directory.
    pseudos = ["Mg-sp-gw.psp8", "O.psp8"]

    # Input for GS part.
    gs_inp = abilab.AbinitInput(structure, pseudos=pseudos)

    gs_inp.set_vars(
        nband=16,
        paral_kgb=0,
        ecut=35.0,        # Too low. Should be ~50
        ngkpt=[4, 4, 4],  # Too coarse
        #nshiftk=1,
        shiftk=[0, 0, 0], # Gamma-centered mesh. Important to have the CBM/VBM!
        tolvrs=1.0e-10,
        diemac=9.0,
        nstep=150,
        nbdbuf=4,
        #prtpot=1,       # Print potential for Sternheimer
        iomode=3,        # Produce output files in netcdf format.
    )

    return gs_inp


def build_flow(options):
    """
    Create a `Flow` for phonon calculations. The flow has two works.

    - work[0]: GS + NSCF along a k-path
    - work[1]: DFPT work with phonons on a 4x4x4, BECS and eps_inf
    """
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_", "flow_")

    flow = flowtk.Flow(workdir=options.workdir)

    # Build input for GS calculation and create first work with 1 ScfTask.
    gs_inp = make_scf_input()
    work = flow.register_scf_task(gs_inp)
    scf_task = work[0]

    # Build new input for NSCF calculation with k-path (automatically selected by AbiPy)
    # Used to plot the KS band structure
    nscf_kpath_inp = gs_inp.new_with_vars(
        nband=16,
        tolwfr=1e-18,
        iscf=-2,
    )
    nscf_kpath_inp.set_kpath(ndivsm=10)
    work.register_nscf_task(nscf_kpath_inp, deps={scf_task: "DEN"})

    # NSCF run with eph_ngkpt k-mesh to produce the WFK file for the EPH code.
    nscf_eph_inp = gs_inp.new_with_vars(
        tolwfr=1e-16,
        iscf=-2,
        prtpot=1,         # Needed for the Sternheimer.
                          # This POT file is actually a copy of the GS POT file.
    )

    eph_ngkpt = [32, 32, 32]
    nscf_eph_inp.set_kmesh(
        ngkpt=eph_ngkpt,
        shiftk=[0.0, 0.0, 0.0],
    )
    work.register_nscf_task(nscf_eph_inp, deps={scf_task: "DEN"})
    nscf_eph_work = work[-1]

    # Q-mesh for phonons. In this case k-mesh == q-mesh
    #ngkpt = [32, 32, 32]
    ngqpt = [4, 4, 4]

    # Create work for phonon calculationwith a [4, 4, 4] q-mesh.
    # Electric field and Born effective charges are also computed.
    ph_work = flowtk.PhononWork.from_scf_task(scf_task, ngqpt, is_ngqpt=True, with_becs=True, with_quad=False)

    flow.register_work(ph_work)

    # Build template for self-energy calculation. See also v8/Input/t44.in
    # The k-points must be in the WFK file
    #
    eph_inp = gs_inp.new_with_vars(
        optdriver=7,             # Enter EPH driver.
        eph_task=4,              # Activate computation of EPH self-energy.
        eph_stern=1,             # Use the sternheimer equations
        ngkpt=eph_ngkpt,
        ddb_ngqpt=ngqpt,         # q-mesh used to produce the DDB file (must be consistent with DDB data)
        symsigma=1,              # Use symmetries in self-energy integration (IBZ_k instead of BZ)
        # For more k-points...
        nkptgw=1,
        kptgw=[0, 0, 0],
        bdgw=[9, 9],
        #gw_qprange=-4,
        nfreqsp=8000,
        freqspmax="8.0 eV",
        tmesh=[0, 200, 1],    # (start, step, num)
        zcut="0.01 eV",
    )

    # Set q-path for Fourier interpolation of phonons.
    eph_inp.set_qpath(10)

    # Set q-mesh for phonons DOS.
    eph_inp.set_phdos_qmesh(nqsmall=16, method="tetra")

    # Convergence of the q-mesh grid
    # For each q-mesh check the frequency mesh and thus the time mesh density to obtain a stable spectral function
    for eph_ngqpt_fine in [ [8, 8, 8], [16, 16, 16], [32, 32, 32]]:
        # Create empty work to contain EPH tasks with this value of eph_ngqpt_fine
        eph_work = flow.register_work(flowtk.Work())

        for nfreq in [1000, 2000, 3000, 4000]:

            # EPH part requires the GS WFK, the DDB file with all perturbations
            # and the database of DFPT potentials (already merged by PhononWork)
            deps = {nscf_eph_work: ["WFK", "POT"], ph_work: ["DDB", "DVDB"]}
            new_inp = eph_inp.new_with_vars(eph_ngqpt_fine=eph_ngqpt_fine, nfreqsp=nfreq)
            eph_work.register_eph_task(new_inp, deps=deps)

            ce_inp = new_inp.new_with_vars(
                eph_task=9,
                tolcum=1e-3
            )
            # Cumulant expansion requires the SIGEPH output file obtained from the EPH calculation
            deps = {nscf_eph_work: ["WFK", "POT"], ph_work: ["DDB", "DVDB"], eph_work[-1]: "SIGEPH"}

            eph_work.register_eph_task(ce_inp, deps=deps)

    # Create workflow for the EPH task
    #eph_work.register_eph_task(eph_inp, deps=deps)

    #ce_inp = eph_inp.new_with_vars(
    #        eph_task=9,
    #        tolcum=1e-3
    #        )

    #deps = {nscf_eph_work: ["WFK","POT"], ph_work: ["DDB", "DVDB"], eph_work[0]: "SIGEPH"}

    #eph_work.register_eph_task(ce_inp, deps=deps)

    flow.allocate()
    flow.use_smartio()

    return flow

# This block generates the thumbnails in the AbiPy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())





"""Panels to interact with A2F files."""
from __future__ import annotations

#import param
import panel as pn
import panel.widgets as pnw

from .core import (PanelWithElectronBands, ply, mpl, dfc, depends_on_btn_click)


class A2fFilePanel(PanelWithElectronBands):
    """
    Panel with widgets to interact with a |GsrFile|.
    """
    def __init__(self, ncfile, **params):
        PanelWithElectronBands.__init__(self, ebands=ncfile.ebands, **params)
        self.ncfile = ncfile

        self.a2f_view_btn = pnw.Button(name="Plot a2F", button_type='primary')

    def get_a2f_view(self) -> pn.Row:
        """
        Return Row with widgets to visualize the structure.
        """
        return pn.Row(
            self.pws_col(["## Visualize a2F",
                          "a2f_view_btn",
                          ]),
            self.on_view_a2f,
        )

    @depends_on_btn_click('a2f_view_btn')
    def on_view_a2f(self):
        """Visualize a2f function."""
        col = pn.Column(sizing_mode='stretch_width'); ca = col.append

        ncfile = self.ncfile

        a2f = ncfile.a2f_qcoarse
        ca(f"## A2F")
        fig = a2f.plot_with_lambda(units="eV", **self.mpl_kwargs)
        ca(mpl(fig))

        ca(f"## A2F + nu")
        fig = a2f.plot_nuterms(units="eV", ax_mat=None, with_lambda=True, fontsize=12,
                               xlims=None, ylims=None, label=None, **self.mpl_kwargs)
        ca(mpl(fig))

        for what in ("gamma", "lambda"):
            ca(f"## Phbands with {what}")
            fig = ncfile.plot(what=what, units="eV", scale=None, alpha=0.6, ylims=None,
                              ax=None, colormap="jet", **self.mpl_kwargs)
            ca(mpl(fig))

        return col

    def get_panel(self, as_dict=False, **kwargs):
        """
        Return tabs with widgets to interact with the A2F file.
        """
        d = {}

        #d["Summary"] = self.get_summary_view_for_abiobj(self.gsr)
        d["a2F"] = self.get_a2f_view()
        d["e-Bands"] = self.get_plot_ebands_view()

        #tabs_dict = self.ncfile.phbands.get_panel(as_dict=True)
        #for k, v in tabs_dict.items():
        #    d[k] = v

        kpoints = self.ncfile.ebands.kpoints
        if kpoints.is_ibz:
            # Add DOS tab but only if k-sampling.
            d["e-DOS"] = self.get_plot_edos_view()
            d["SKW"] = self.get_skw_view()

            if not self.ncfile.ebands.isnot_ibz_sampling():
                d["Ifermi"] = self.get_ifermi_view()
                #d["fsviewer"] = self.get_fsviewer_view()

        d["Structure"] = self.get_structure_view()
        d["NcFile"] = self.ncfile.get_ncfile_view()

        # TODO
        #d["Global"] = pn.Row(
        #    pn.Column("# Global options",
        #              *self.pws("units", "mpi_procs", "verbose"),
        #              ),
        #    self.get_software_stack())
        #))

        if as_dict: return d
        return self.get_template_from_tabs(d, template=kwargs.get("template", None))



#!/usr/bin/env python
"""
This script regenerate all the reference files located in this directory 
"""
from __future__ import print_function, division, unicode_literals

import sys
import os 
import argparse
import shutil
import tempfile

from subprocess import call, Popen


def main():
    def str_examples():
        examples = """
          Usage example:\n\n
          runall.py               => Regenerate all reference files.
        """
        return examples

    def show_examples_and_exit(err_msg=None, error_code=1):
        """Display the usage of the script."""
        sys.stderr.write(str_examples())
        if err_msg: 
            sys.stderr.write("Fatal Error\n" + err_msg + "\n")
        sys.exit(error_code)

    parser = argparse.ArgumentParser(epilog=str_examples(),formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('-m', '--mode', type=str, default="sequential",
                        help="execution mode. Default is sequential.")

    parser.add_argument('-e', '--exclude', type=str, default="", help="Exclude scripts.")

    parser.add_argument('--keep-dirs', action="store_true", default=False,
                        help="Do not remove flowdirectories.")

    parser.add_argument('-b', '--bail-on-failure', default=False, help="Exit at the first error.")

    #parser.add_argument("scripts", nargs="+",help="List of scripts to be executed")

    options = parser.parse_args()

    # Find scripts.
    if options.exclude:
        options.exclude = options.exclude.split()
        print("Will exclude:\n", options.exclude)

    scripts = []
    for dirpath, dirnames, filenames in os.walk(os.path.dirname(__file__)):
        for fname in filenames:
            if fname in options.exclude: continue
            if fname == "gendata.py":
                path = os.path.join(os.path.abspath(dirpath), fname)
                #if path != __file__:
                scripts.append(path)

    #print(scripts)
    #return 0

    # Run scripts according to mode.
    dirpaths, retcode = [], 0
    if options.mode in ["s", "sequential"]:
        for script in scripts:
            os.chdir(os.path.dirname(script))
            ret = call(["python", script])
            retcode += ret

            if ret != 0: 
                print("retcode %d while running %s" % (ret, script))
                if options.bail_on_failure: break

        # Remove directories.
        #if not options.keep_dirs:
        #    for dirpath in dirpaths:
        #        try:
        #            shutil.rmtree(dirpath, ignore_errors=False)
        #        except OSError:
        #            print("Exception while removing %s" % dirpath)

    else:
        show_examples_and_exit(err_msg="Wrong value for mode: %s" % options.mode)

    print("retcode %d" % retcode)
    return retcode

if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
"""
Interface to the database of ABINIT input variables
"""
from __future__ import annotations

import sys
import argparse
import abipy.flowtk as flowtk

from monty.functools import prof_main
from monty.termcolor import cprint
from abipy.core.release import __version__
from abipy import abilab


def print_vlist(vlist, options):
    for v in vlist:
        print(repr(v))

    if options.verbose:
        for v in vlist: abilab.abinit_help(v)
    else:
        print("\nUse -v for more info")


def get_epilog() -> str:
    return """\
Usage example:

    abidoc.py man ecut        --> Show documentation for ecut input variable.
    abidoc.py graphviz acell  --> Draw parents and children of variable with graphviz package.
    abidoc.py browse acell    --> Open url in external browser.
    abidoc.py apropos ecut    --> To search in the database for the variables related to ecut.
    abidoc.py find paw        --> To search in the database for the variables whose name contains paw.
    abidoc.py list            --> Print full list of variables.
    abidoc.py withdim natom   --> Print arrays depending on natom.
    abidoc.py scheduler       --> Document scheduler options.
    abidoc.py manager         --> Document manager options.

Use `abidoc.py --help` for help and `abidoc.py COMMAND --help` to get the documentation for `COMMAND`.
Use `-v` to increase verbosity level (can be supplied multiple times e.g -vv).
"""


def get_parser(with_epilog=False):

    # Build the main parser.
    parser = argparse.ArgumentParser(epilog=get_epilog() if with_epilog else "",
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('-V', '--version', action='version', version=__version__)

    # Parent parser for common options.
    copts_parser = argparse.ArgumentParser(add_help=False)
    copts_parser.add_argument('-v', '--verbose', default=0, action='count', # -vv --> verbose=2
                              help='verbose, can be supplied multiple times to increase verbosity')
    copts_parser.add_argument('--loglevel', default="ERROR", type=str,
                              help="Set the loglevel. Possible values: CRITICAL, ERROR (default), WARNING, INFO, DEBUG")
    copts_parser.add_argument("-c", '--codename', type=str, default="abinit",
                              help="Code name e.g. anaddb, optic... Default: abinit")

    var_parser = argparse.ArgumentParser(add_help=False)
    var_parser.add_argument('varname', help="ABINIT variable")

    # Create the parsers for the sub-commands
    subparsers = parser.add_subparsers(dest='command', help='sub-command help', description="Valid subcommands")

    # Subparser for man.
    p_man = subparsers.add_parser('man', parents=[copts_parser, var_parser], help="Show documentation for varname.")

    # Subparser for graphviz.
    p_graphviz = subparsers.add_parser('graphviz', parents=[copts_parser, var_parser],
        help=("Draw variable dependencies with graphviz package."
              "See https://graphviz.readthedocs.io/."))
    p_graphviz.add_argument("-e", "--engine", type=str, default="automatic",
        help=("graphviz engine: ['dot', 'neato', 'twopi', 'circo', 'fdp', 'sfdp', 'patchwork', 'osage']. "
            "Default: automatic i.e. the engine is automatically selected. See http://www.graphviz.org/pdf/dot.1.pdf "
            "Use `conda install python-graphviz` or `pip install graphviz` to install the python package"))

    # Subparser for browse.
    p_browse = subparsers.add_parser('browse', parents=[copts_parser, var_parser], help="Open documentation in browser.")

    # Subparser for apropos.
    p_apropos = subparsers.add_parser('apropos', parents=[copts_parser, var_parser],
                                      help="Find variables related to varname.")
    # Subparser for find.
    p_find = subparsers.add_parser('find', parents=[copts_parser, var_parser],
                                   help="Find all variables whose name contains varname.")
    # Subparser for require.
    #p_require = subparsers.add_parser('require', parents=[copts_parser], help="Find all variables required by varname.")

    # Subparser for withdim.
    p_withdim = subparsers.add_parser('withdim', parents=[copts_parser],
                                      help="Find all arrays depending on the given dimension.")
    p_withdim.add_argument("dimname", help="Dimension name")

    # Subparser for list.
    p_list = subparsers.add_parser('list', parents=[copts_parser], help="List all variables.")
    p_list.add_argument('--mode', default="a",
                        help="Sort mode, `a` for alphabethical, `s` for varset, `c` for characteristics.")

    # Subparser for manager.
    p_manager = subparsers.add_parser('manager', parents=[copts_parser], help="Document the TaskManager options.")
    p_manager.add_argument("qtype", nargs="?", default=None, help=("Write job script to terminal if qtype='script' else "
        "document the qparams for the given QueueAdapter qtype e.g. slurm."))

    # Subparser for scheduler
    p_docsched = subparsers.add_parser('scheduler', parents=[copts_parser],
        help="Document the options available in scheduler.yml.")

    # Subparser for abibuild
    p_abibuild = subparsers.add_parser('abibuild', parents=[copts_parser],
        help="Show ABINIT build information and exit.")

    # Subparser for links
    p_urls = subparsers.add_parser('urls', parents=[copts_parser],
        help="Produce Yaml file with varname --> ulrs.")

    return parser


@prof_main
def main():

    def show_examples_and_exit(err_msg=None, error_code=1):
        """Display the usage of the script."""
        sys.stderr.write(get_epilog())
        if err_msg: sys.stderr.write("Fatal Error\n" + err_msg + "\n")
        sys.exit(error_code)

    parser = get_parser(with_epilog=True)

    try:
        options = parser.parse_args()
    except Exception as exc:
        show_examples_and_exit(error_code=1)

    # loglevel is bound to the string value obtained from the command line argument.
    # Convert to upper case to allow the user to specify --loglevel=DEBUG or --loglevel=debug
    import logging
    numeric_level = getattr(logging, options.loglevel.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError('Invalid log level: %s' % options.loglevel)
    logging.basicConfig(level=numeric_level)

    # Get the dabase of variables for codename.
    from abipy.abio.abivar_database.variables import get_codevars
    codevars = get_codevars()
    vdb = codevars[options.codename]

    if options.command == "man":
        abilab.abinit_help(options.varname)

    elif options.command == "browse":
        return vdb[options.varname].browse()

    elif options.command == "graphviz":
        if options.varname in vdb.my_varset_list:
            graph = vdb.get_graphviz(varset=options.varname, vartype=None, engine=options.engine)
        else:
            graph = vdb.get_graphviz_varname(varname=options.varname, engine=options.engine)

        import tempfile
        directory = tempfile.mkdtemp()
        print("Producing source files in:", directory)
        graph.view(directory=directory, cleanup=False)

    elif options.command == "apropos":
        vlist = vdb.apropos(options.varname)
        print_vlist(vlist, options)

    elif options.command == "find":
        vlist = [v for v in vdb.values() if options.varname in v.name]
        print("Find results:\n")
        print_vlist(vlist, options)

    elif options.command == "list":
        if options.mode == "a":
            # Alphabetical
            for i, var in enumerate(vdb.values()):
                print(i, repr(var))

        elif options.mode == "s":
            # Grouped by varset
            for section in vdb.my_varset_list:
                print(30 * "#" + " Section: " + section + " " + 30 * "#")
                print_vlist(vdb.vars_with_section(section), options)

        elif options.mode == "c":
            # Grouped by characteristics.
            for char in vdb.my_characteristics:
                print(30 * "#" + " Characteristic: " + char + 30 * "#")
                print_vlist(vdb.vars_with_char(char), options)

        else:
            raise ValueError("Wrong mode %s" % options.mode)

    elif options.command == "withdim":
        for var in vdb.values():
            if var.depends_on_dimension(options.dimname):
                cprint(repr(var), "yellow")
                print("dimensions:", str(var.dimensions), "\n")

    elif options.command == "scheduler":
        print("Options that can be specified in scheduler.yml:")
        print(flowtk.PyFlowScheduler.autodoc())
        return 0

    elif options.command == "manager":
        # Document TaskManager options and qparams.
        qtype = options.qtype

        if qtype == "script":
            manager = flowtk.TaskManager.from_user_config()
            script = manager.qadapter.get_script_str(
                job_name="job_name", launch_dir="workdir", executable="executable",
                qout_path="qout_file.path", qerr_path="qerr_file.path",
                stdin="stdin", stdout="stdout", stderr="stderr")
            print(script)

        else:
            print(flowtk.TaskManager.autodoc())
            print("qtype supported: %s" % flowtk.all_qtypes())
            print("Use:\n\n\tabidoc.py manager slurm\n\nto show the list of qparams for slurm.\n")

            if qtype is not None:
                print("QPARAMS for %s" % qtype)
                flowtk.show_qparams(qtype)

    elif options.command == "abibuild":
        abinit_build = flowtk.AbinitBuild()
        print()
        print(abinit_build)
        print()
        if not options.verbose:
            print("Use --verbose for additional info")
        else:
            print(abinit_build.info)

    elif options.command == "urls":
        d = {}
        for codename in sorted(codevars.keys()):
            variables = codevars[codename]
            for varname, var in variables.items():
                key = var.abivarname
                key.replace("@", "_")
                d[key] = f"[{var.name}]({var.website_url})"
        from ruamel.yaml import YAML
        YAML().dump(d, sys.stdout)

    else:
        raise ValueError("Don't know how to handle command %s" % options.command)

    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
r"""
Flow for computing e-ph matrix elements along a q-path
======================================================

This example shows how to compute the phonon band structure of AlAs with AbiPy flows.
The final results (out_DDB, out_DVDB) will be produced automatically at the end of the run
and saved in ``flow_phonons/outdata/``.
"""
from __future__ import division, print_function, unicode_literals, absolute_import

import sys
import os
import abipy.abilab as abilab
import abipy.data as abidata
import numpy as np

from abipy import flowtk

def make_scf_input(ngkpt):
    """
    This function constructs the input file for the GS calculation:
    """
    structure = dict(
     angdeg= 3*[60.0],
     acell=3* [ 7.1992351952],
     natom=2,
     ntypat=2,
     typat=[1, 2],
     znucl=[31, 15],
     xred=[
        0.0000000000, 0.0000000000, 0.0000000000,
        0.2500000000, 0.2500000000, 0.2500000000,
    ])

    #pseudos = ["Ga.psp8", "P.psp8"]
    pseudos = abidata.pseudos("Ga.oncvpsp", "P.psp8")
    gs_inp = abilab.AbinitInput(structure, pseudos=pseudos)

    gs_inp.set_vars(
        nband=8,
        #nband=13,
        ecut=30.0,
        #ecut=24.0,
        ngkpt=ngkpt,
        nshiftk=1,
        shiftk=[0, 0, 0],
        tolvrs=1.0e-10,
        nstep=150,
    )

    return gs_inp


def build_flow(options):
    """
    Create a `Flow` for phonon calculations. The flow has two works.

    The first work contains a single GS task that produces the WFK file used in DFPT
    Then we have multiple Works that are generated automatically
    in order to compute the dynamical matrix on a [2, 2, 2] mesh.
    Symmetries are taken into account: only q-points in the IBZ are generated and
    for each q-point only the independent atomic perturbations are computed.
    """
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(__file__).replace(".py", "").replace("run_", "flow_")

    # Use 2x2x2 both for k-mesh and q-mesh
    # Build input for GS calculation
    scf_input = make_scf_input(ngkpt=(4, 4, 4))

    # Create flow to compute all the independent atomic perturbations
    # corresponding to a [4, 4, 4] q-mesh.
    # Electric field and Born effective charges are also computed.
    from abipy.flowtk.eph_flows import EphPotFlow
    ngqpt = (2, 2, 2)
    #ngqpt = (4, 4, 4)
    #qpath_list = [[0.01, 0, 0], [0.02, 0, 0], [0.24, 0, 0], [0.45, 0, 0]]
    #qpath_list = [[0.0, 0.0, 0.0], [0.01, 0, 0], [0.1, 0, 0],
    #              [0.24, 0, 0], [0.3, 0, 0], [0.45, 0, 0], [0.5, 0.0, 0.0]]

    #qpath_list = [[0.0, 0.0, 0.0], [0.01, 0, 0], [0.1, 0, 0],
    #              [0.24, 0, 0], [0.3, 0, 0], [0.45, 0, 0], [0.5, 0.0, 0.0]]

    qpath_list = [
    #+0.50000,  +0.50000,  +0.50000,  # L
    #+0.00000,  +0.00000,  +0.00000,  # $\Gamma$
    #+0.50000,  +0.00000,  +0.50000,  # X

    +0.10000,  +0.10000,  +0.10000,  # L
    +0.00000,  +0.00000,  +0.00000,  # $\Gamma$
    +0.10000,  +0.00000,  +0.10000,  # X

    #+0.50000,  +0.25000,  +0.75000,  # W
    #+0.37500,  +0.37500,  +0.75000,  # K
    #+0.00000,  +0.00000,  +0.00000,  # $\Gamma$
    #+0.50000,  +0.25000,  +0.75000,  # W

    #+0.62500  +0.25000  +0.62500  # U
    #+0.50000  +0.25000  +0.75000  # W
    #+0.50000  +0.50000  +0.50000  # L
    #+0.37500  +0.37500  +0.75000  # K
    #+0.62500  +0.25000  +0.62500  # U
    #+0.50000  +0.00000  +0.50000  # X
    ]

    flow = EphPotFlow.from_scf_input(options.workdir, scf_input,
                                  ngqpt, qpath_list, ndivsm=5, ddk_tolerance={"tolwfr": 1e-12},
                                  with_becs=True) #, prepgkk=0)

    return flow


# This block generates the thumbnails in the Abipy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())





# coding: utf-8
"""
RTA.nc file.
"""
from __future__ import annotations

import numpy as np
import abipy.core.abinit_units as abu

from monty.functools import lazy_property
#from monty.termcolor import cprint
from monty.string import marquee, list_strings
from abipy.core.structure import Structure
from abipy.core.mixins import AbinitNcFile, Has_Structure, Has_ElectronBands, NotebookWriter
from abipy.electrons.ebands import ElectronBands, ElectronsReader, RobotWithEbands
from abipy.tools.plotting import add_fig_kwargs, get_ax_fig_plt, get_axarray_fig_plt
from abipy.tools.typing import Figure
from abipy.abio.robots import Robot


__all__ = [
    "RtaFile",
    "RtaRobot",
]


def eh2s(eh) -> str:
    return {0: "n", 1: "p"}[eh]


def irta2s(irta) -> str:
    """Return RTA type from irta index."""
    return {0: "SERTA", 1: "MRTA"}[irta]


def style_for_irta(irta, with_marker=False) -> dict:
    """
    Return dict with linestyle to plot SERTA/MRTA results
    """
    if irta == 0:
        opts = dict(linewidth=1.0, linestyle="dotted")
        if with_marker: opts["marker"] = "^"
    elif irta == 1:
        opts = dict(linewidth=1.0, linestyle="solid")
        if with_marker: opts["marker"] = "v"
    else:
        raise ValueError("Invalid value for irta: %s" % irta)

    return opts


def transptens2latex(what, component) -> str:
    return {
        "sigma": r"$\sigma_{%s}$" % component,
        "seebeck": "$S_{%s}$" % component,
        "kappa": r"$\kappa^{\mathcal{e}}_{%s}$" % component,
        "pi": r"$\Pi_{%s}$" % component,
        "zte": r"$\text{ZT}^e_{%s}$" % component,
    }[what]


def edos_infos(edos_intmeth, edos_broad) -> str:
    s = {1: "Gaussian smearing Method",
         2: "Linear Tetrahedron Method",
        -2: "Linear Tetrahedron Method with Blochl's corrections",
    }[edos_intmeth]
    if (edos_intmeth == 1): s = "%s with broadening: %.1f (meV)" % edos_broad * abu.Ha_to_meV

    return s


def irta2latextau(irta, with_dollars=False) -> str:
    s = r"\tau^{\mathbf{%s}}}" % irta2s(irta)
    if with_dollars: s = "$%s$" % s
    return s


def x2_grid(what_list) -> tuple:
    """
    Build (x, 2) grid of plots or just (1, 1) depending of the length of what_list.

    Return: (num_plots, ncols, nrows, what_list)
    """
    what_list = list_strings(what_list)
    num_plots, ncols, nrows = len(what_list), 1, 1
    if num_plots > 1:
        ncols = 2
        nrows = (num_plots // ncols) + (num_plots % ncols)

    return num_plots, ncols, nrows, what_list


class RtaFile(AbinitNcFile, Has_Structure, Has_ElectronBands, NotebookWriter):

    @classmethod
    def from_file(cls, filepath: str) -> RtaFile:
        """Initialize the object from a netcdf file."""
        return cls(filepath)

    def __init__(self, filepath: str):
        super().__init__(filepath)
        self.reader = RtaReader(filepath)

        self.nrta = self.reader.read_dimvalue("nrta")

        #self.fermi = self.ebands.fermie * abu.eV_Ha
        #self.transport_ngkpt = self.reader.read_value("transport_ngkpt")
        #self.transport_extrael = self.reader.read_value("transport_extrael")
        #self.transport_fermie = self.reader.read_value("transport_fermie")
        self.sigma_erange = self.reader.read_value("sigma_erange")
        #self.ebands.kpoints.ksampling.mpdivs

        # Get position of CBM and VBM for each spin in eV
        # nctkarr_t('vb_max', "dp", "nsppol")
        self.vb_max_spin = self.reader.read_value("vb_max") * abu.Ha_to_eV
        self.cb_min_spin = self.reader.read_value("cb_min") * abu.Ha_to_eV

        # Get metadata for k-integration (coming from edos%ncwrite)
        self.edos_intmeth = int(self.reader.read_value("edos_intmeth"))
        self.edos_broad = self.reader.read_value("edos_broad")

        # Store also the e-mesh n eV as it's often needed in the plotting routines.
        # Several quantities are defined on this mesh.
        self.edos_mesh_eV = self.reader.read_value("edos_mesh") * abu.Ha_to_eV

    @property
    def ntemp(self) -> int:
        """Number of temperatures."""
        return len(self.tmesh)

    @property
    def tmesh(self) -> np.ndarray:
        """Mesh with Temperatures in Kelvin."""
        return self.reader.tmesh

    @lazy_property
    def assume_gap(self) -> bool:
        """True if we are dealing with a semiconductor. More precisely if all(sigma_erange) > 0."""
        return bool(self.reader.rootgrp.variables["assume_gap"])

    @lazy_property
    def has_ibte(self) -> bool:
        """True if file contains IBTE results."""
        return "ibte_sigma" in self.reader.rootgrp.variables

    @lazy_property
    def ebands(self) -> ElectronBands:
        """|ElectronBands| object."""
        return self.reader.read_ebands()

    @property
    def structure(self) -> Structure:
        """|Structure| object."""
        return self.ebands.structure

    @lazy_property
    def params(self) -> dict:
        """dict with parameters that might be subject to convergence studies."""
        od = self.get_ebands_params()
        return od

    def __str__(self) -> str:
        """String representation."""
        return self.to_string()

    def to_string(self, verbose=0) -> str:
        """String representation."""
        lines = []; app = lines.append

        app(marquee("File Info", mark="="))
        app(self.filestat(as_string=True))
        app("")
        app(self.structure.to_string(verbose=verbose, title="Structure"))
        app("")
        app(self.ebands.to_string(with_structure=False, verbose=verbose, title="KS Electron Bands"))
        app("")

        # Transport section.
        app(marquee("Transport calculation", mark="="))
        app("")
        #app("edos_intmeth: %d" % self.edos_intmeth)
        #app("edos_broad: %d (meV): " % (self.edos_broad * 1000))
        app(edos_infos(self.edos_intmeth, self.edos_broad))
        app("mesh step for energy integrals: %.1f (meV) " % ((self.edos_mesh_eV[1] - self.edos_mesh_eV[0]) * 1000))
        app("")

        components = ("xx", "yy", "zz") if verbose == 0 else ("xx", "yy", "zz", "xy", "xz", "yx")
        for component in components:
            for irta in range(self.nrta):
                app("Mobility (%s Cartesian components), RTA type: %s" % (component, irta2s(irta)))
                app("Temperature [K]     Electrons (cm^2/Vs)     Holes (cm^2/Vs)")
                for itemp in range(self.ntemp):
                    temp = self.tmesh[itemp]
                    mobility_mu_e = self.get_mobility_mu(eh=0, itemp=itemp, component=component, irta=irta)
                    mobility_mu_h = self.get_mobility_mu(eh=1, itemp=itemp, component=component, irta=irta)
                    app("%14.1lf %18.6lf %18.6lf" % (temp, mobility_mu_e, mobility_mu_h))
                app("")

        return "\n".join(lines)

    def get_mobility_mu(self, eh, itemp, component='xx', ef=None, irta=0, spin=0):
        """
        Get the mobility at the chemical potential Ef

        Args:
            eh: 0 for electrons, 1 for holes.
            itemp: Index of the temperature.
            component: Cartesian component to plot: "xx", "yy" "xy" ...
            ef: Value of the doping in eV.
                The default None uses the chemical potential at the temperature item as computed by Abinit.
            spin: Spin index.
        """
        if ef is None: ef = self.reader.read_value('transport_mu_e')[itemp]
        emesh, mobility = self.reader.read_mobility(eh, itemp, component, spin, irta=irta)

        from scipy import interpolate
        f = interpolate.interp1d(emesh, mobility)
        return f(ef)

    #def get_mobility_mu_dataframe(self, eh=0, component='xx', itemp=0, spin=0, **kwargs):

    #def _select_itemps_labels(self, obj):
    #   for it, temp in enumerate(self.tmesh):

    def _add_vline_at_bandedge(self, ax, spin, cbm_or_vbm, **kwargs):
        my_kwargs = dict(ymin=0, ymax=1, linewidth=1, linestyle="--")
        my_kwargs.update(kwargs)
        #from matplotlib.pyplot import text

        if cbm_or_vbm in ("cbm", "both"):
            x = self.cb_min_spin[spin]
            ax.axvline(x=x, color="red", **my_kwargs) # label="CBM",
            #ax.text(x, 5, "CBM", rotation=90, verticalalignment='center', fontsize=8)

        if cbm_or_vbm in ("vbm", "both"):
            x = self.vb_max_spin[spin]
            ax.axvline(x=x, color="blue", **my_kwargs) # label="VBM",
            #ax.text(x, 5, "VBM", rotation=90, verticalalignment='center', fontsize=8)

    @add_fig_kwargs
    def plot_edos(self, ax=None, fontsize=8, **kwargs) -> Figure:
        """
        Plot electron DOS

        Args:
            ax: |matplotlib-Axes| or None if a new figure should be created.
            fontsize (int): fontsize for titles and legend

        Return: |matplotlib-Figure|
        """
        ax, fig, plt = get_ax_fig_plt(ax=ax)

        # Total DOS, spin up and spin down components in nsppol_plus1.
        # nctkarr_t("edos_dos", "dp", "edos_nw, nsppol_plus1")
        dos = self.reader.read_value("edos_dos") / abu.Ha_to_eV

        # Plot total DOS.
        ax.plot(self.edos_mesh_eV, dos[0], label="Total DOS", color="black", linewidth=1.0)

        #idos = self.reader.read_value("edos_idos")
        #ax.plot(self.edos_mesh_eV, idos[0], label="Total IDOS", color="black", linewidth=1.0)

        if self.nsppol == 2:
            ax.plot(self.edos_mesh_eV, + dos[1], color="red", linewidth=1, label="up")
            ax.plot(self.edos_mesh_eV, - dos[2], color="blue", linewidth=1, label="down")

        for spin in range(self.nsppol):
            self._add_vline_at_bandedge(ax, spin, "both")

        ax.grid(True)
        ax.set_xlabel('Energy (eV)')
        ax.set_ylabel('States/eV p.u.c')
        ax.legend(loc="best", shadow=True, fontsize=fontsize)

        if "title" not in kwargs:
            title = r"$\frac{1}{N_k} \sum_{nk} \delta(\epsilon - \epsilon_{nk})$"
            fig.suptitle(title, fontsize=fontsize)

        return fig

    @add_fig_kwargs
    def plot_tau_isoe(self, ax_list=None, colormap="jet", fontsize=8, **kwargs) -> Figure:
        r"""
        Plot tau(e). Energy-dependent scattering rate defined by:

            $\tau(\epsilon) = \frac{1}{N_k} \sum_{nk} \tau_{nk}\,\delta(\epsilon - \epsilon_{nk})$

        Two differet subplots for SERTA and MRTA.

        Args:
            ax_list: List of |matplotlib-Axes| or None if a new figure should be created.
            colormap:
            fontsize (int): fontsize for titles and legend

        Return: |matplotlib-Figure|
        """
        ax_list, fig, plt = get_axarray_fig_plt(ax_list, nrows=self.nrta, ncols=1,
                                                sharex=True, sharey=True, squeeze=False)
        ax_list = ax_list.ravel()
        cmap = plt.get_cmap(colormap)

        # nctkarr_t('tau_dos', "dp", "edos_nw, ntemp, nsppol, nrta")
        tau_dos = self.reader.read_value("tau_dos")

        for irta, ax in enumerate(ax_list):
            for spin in range(self.nsppol):
                spin_sign = +1 if spin == 0 else -1
                for it, temp in enumerate(self.tmesh):
                    # Convert to femtoseconds
                    ys = spin_sign * tau_dos[irta, spin, it] * abu.Time_Sec * 1e+15
                    ax.plot(self.edos_mesh_eV , ys, c=cmap(it / self.ntemp),
                            label="T = %dK" % temp if spin == 0 else None)

            ax.grid(True)
            ax.legend(loc="best", shadow=True, fontsize=fontsize)
            if irta == (len(ax_list) - 1):
                ax.set_xlabel('Energy (eV)')
                ax.set_ylabel(r"$\tau(\epsilon)\, (fms)$")

            self._add_vline_at_bandedge(ax, spin, "both")

            ax.text(0.1, 0.9, irta2s(irta), fontsize=fontsize,
                horizontalalignment='center', verticalalignment='center', transform=ax.transAxes,
                bbox=dict(alpha=0.5))

        if "title" not in kwargs:
            title = r"$\tau(\epsilon) = \frac{1}{N_k} \sum_{nk} \tau_{nk}\,\delta(\epsilon - \epsilon_{nk})$"
            fig.suptitle(title, fontsize=fontsize)

        return fig

    #@add_fig_kwargs
    #def plot_vv_dos(self, component="xx", spin=0, ax=None, fontsize=8, **kwargs):

    @add_fig_kwargs
    def plot_vvtau_dos(self, component="xx", spin=0, ax=None, colormap="jet", fontsize=8, **kwargs) -> Figure:
        r"""
        Plot (v_i * v_j * tau) DOS.

            $\frac{1}{N_k} \sum_{nk} v_i v_j \delta(\epsilon - \epsilon_{nk})$

        Args:
            component: Cartesian component to plot: "xx", "yy" "xy" ...
            ax: |matplotlib-Axes| or None if a new figure should be created.
            colormap: matplotlib colormap.
            fontsize (int): fontsize for titles and legend

        Return: |matplotlib-Figure|
        """
        i, j = abu.s2itup(component)

        ax, fig, plt = get_ax_fig_plt(ax=ax)
        cmap = plt.get_cmap(colormap)

        for irta in range(self.nrta):
            # nctkarr_t('vvtau_dos', "dp", "edos_nw, three, three, ntemp, nsppol, nrta")
            var = self.reader.read_variable("vvtau_dos")
            for itemp, temp in enumerate(self.tmesh):
                vvtau_dos = var[irta, spin, itemp, j, i, :] / (2 * abu.Ha_s)
                label = "T = %dK" % temp
                if (itemp == 0): label = "%s (%s)" % (label, irta2s(irta))
                if (irta == 0 and itemp > 0): label = None
                ax.plot(self.edos_mesh_eV, vvtau_dos, c=cmap(itemp / self.ntemp), label=label, **style_for_irta(irta))

                # This to plot the vv dos along without tau
                #if itemp == 1:
                #    # nctkarr_t('vv_dos', "dp", "edos_nw, three, three, nsppol"), &
                #    vv_dos_var = self.reader.read_variable("vv_dos")
                #    vv_dos = vv_dos_var[spin, j, i] # / (2 * abu.Ha_s)
                #    ax.plot(self.edos_mesh_eV, vv_dos, c=cmap(itemp / self.ntemp), label='VVDOS' % temp)

        self._add_vline_at_bandedge(ax, spin, "both")

        ax.grid(True)
        ax.set_xlabel('Energy (eV)')
        ax.set_ylabel(r'$v_{%s} v_{%s} \tau$ DOS' % (component[0], component[1]))
        ax.set_yscale('log')
        ax.legend(loc="best", shadow=True, fontsize=fontsize)

        if "title" not in kwargs:
            vvt = r'v_{%s} v_{%s} \tau' % (component[0], component[1])
            title = r"$\frac{1}{N_k} \sum_{nk} %s\,\delta(\epsilon - \epsilon_{nk})$" % vvt
            fig.suptitle(title, fontsize=fontsize)

        return fig

    @add_fig_kwargs
    def plot_mobility(self, eh=0, irta=0, component='xx', spin=0, ax=None,
                      colormap='jet', fontsize=8, yscale="log", **kwargs) -> Figure:
        """
        Read the mobility from the netcdf file and plot it

        Args:
            component: Component to plot: "xx", "yy" "xy" ...
            ax: |matplotlib-Axes| or None if a new figure should be created.
            colormap: matplotlib colormap.
            fontsize (int): fontsize for titles and legend

        Return: |matplotlib-Figure|
        """
        ax, fig, plt = get_ax_fig_plt(ax=ax)
        cmap = plt.get_cmap(colormap)

        # nctkarr_t('mobility',"dp", "three, three, edos_nw, ntemp, two, nsppol, nrta")
        mu_var = self.reader.read_variable("mobility")
        i, j = abu.s2itup(component)

        for irta in range(self.nrta):
            for itemp, temp in enumerate(self.tmesh):
                mu = mu_var[irta, spin, eh, itemp, :, j, i]
                label = "T = %dK" % temp
                if (itemp == 0): label = "%s (%s)" % (label, irta2s(irta))
                if (irta == 0 and itemp > 0): label = None
                ax.plot(self.edos_mesh_eV, mu, c=cmap(itemp / self.ntemp), label=label, **style_for_irta(irta))

        self._add_vline_at_bandedge(ax, spin, "cbm" if eh == 0 else "vbm")

        ax.grid(True)
        ax.set_xlabel('Fermi level (eV)')
        ax.set_ylabel(r'%s-mobility $\mu_{%s}(\epsilon_F)$ (cm$^2$/Vs)' % (eh2s(eh), component))
        ax.set_yscale(yscale)
        ax.legend(loc="best", shadow=True, fontsize=fontsize)

        return fig

    @add_fig_kwargs
    def plot_transport_tensors_mu(self, component="xx", spin=0,
                                  what_list=("sigma", "seebeck", "kappa", "pi"),
                                  colormap="jet", fontsize=8, **kwargs) -> Figure:
        """
        Plot selected Cartesian components of transport tensors as a function
        of the chemical potential mu at the given temperature.

        Args:
            ax_list: |matplotlib-Axes| or None if a new figure should be created.
            fontsize: fontsize for legends and titles

        Return: |matplotlib-Figure|
        """
        i, j = abu.s2itup(component)

        num_plots, ncols, nrows, what_list = x2_grid(what_list)
        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                                sharex=True, sharey=False, squeeze=False)
        ax_list = ax_list.ravel()
        # don't show the last ax if numeb is odd.
        if num_plots % ncols != 0: ax_list[-1].axis("off")

        cmap = plt.get_cmap(colormap)

        for iax, (what, ax) in enumerate(zip(what_list, ax_list)):
            irow, icol = divmod(iax, ncols)
            # nctkarr_t('seebeck', "dp", "three, three, edos_nw, ntemp, nsppol, nrta")
            what_var = self.reader.read_variable(what)

            for irta in range(self.nrta):
                for itemp, temp in enumerate(self.tmesh):
                    ys = what_var[irta, spin, itemp, :, j, i]
                    label = "T = %dK" % temp
                    if itemp == 0: label = "%s (%s)" % (label, irta2s(irta))
                    if irta == 0 and itemp > 0: label = None
                    ax.plot(self.edos_mesh_eV, ys, c=cmap(itemp / self.ntemp), label=label, **style_for_irta(irta))

            ax.grid(True)
            ax.set_ylabel(transptens2latex(what, component))

            ax.legend(loc="best", fontsize=fontsize, shadow=True)
            if irow == nrows - 1:
                ax.set_xlabel(r"$\mu$ (eV)")

            self._add_vline_at_bandedge(ax, spin, "both")

        if "title" not in kwargs:
            fig.suptitle("Transport tensors", fontsize=fontsize)

        return fig

    @add_fig_kwargs
    def plot_ibte_vs_rta_rho(self, component="xx", fontsize=8, ax=None, **kwargs) -> Figure:
        """
        Plot resistivity computed with SERTA, MRTA and IBTE
        """
        #if not self.has_ibte:
        #    cprint("Netcdf file does not contain IBTE results", "magenta")
        #    return None

        i = j = 0
        rta_vals = self.reader.read_value("resistivity")
        serta_t = rta_vals[0, :, j, i]
        mrta_t = rta_vals[1, :, j, i]
        ibte_vals = self.reader.read_value("ibte_rho")
        ibte_t = ibte_vals[:, j, i]

        ax, fig, plt = get_ax_fig_plt(ax=ax)
        ax.grid(True)
        ax.plot(self.tmesh, serta_t, label="serta")
        ax.plot(self.tmesh, mrta_t, label="mrta")
        ax.plot(self.tmesh, ibte_t, label="ibte")
        ax.set_xlabel("Temperature (K)")
        ax.set_ylabel(r"Resistivity ($\mu\Omega\;cm$)")
        ax.legend(loc="best", shadow=True, fontsize=fontsize)

        return fig

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        Return figures plotting the transport data
        """
        yield self.plot_ibte_vs_rta_rho(show=False)
        #yield self.plot_tau_isoe(show=False)
        #yield self.plot_transport_tensors_mu(show=False)
        #yield self.plot_edos(show=False)
        #yield self.plot_vvtau_dos(show=False)
        #yield self.plot_mobility(show=False, title="Mobility")
        #if self.has_ibte:
        #    yield self.plot_ibte_vs_rta_rho(show=False)

    def close(self) -> None:
        """Close the file."""
        self.reader.close()

    def write_notebook(self, nbpath=None) -> str:
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        nb.cells.extend([
            nbv.new_code_cell("ncfile = abilab.abiopen('%s')" % self.filepath),
            nbv.new_code_cell('components = ("xx", "yy", "zz", "xy", "xz", "yx")'),
            nbv.new_code_cell("print(ncfile)"),
            nbv.new_code_cell("ncfile.plot_edos();"),
            nbv.new_code_cell("ncfile.plot_vvtau_dos();"),
            nbv.new_code_cell("""
for component in components:
    ncfile.plot_mobility(component=component, spin=0);
"""),
        ])

        return self._write_nb_nbpath(nb, nbpath)


class RtaReader(ElectronsReader):
    """
    This class reads the results stored in the RTA.nc file
    It provides helper function to access the most important quantities.
    """
    def __init__(self, filepath: str):
        super().__init__(filepath)

        self.nsppol = self.read_dimvalue('nsppol')
        self.tmesh = self.read_value("kTmesh") / abu.kb_HaK

    #def read_vvdos_tau(self, itemp, component='xx', spin=0, irta=0):
    #    """
    #    Read the group velocity density of states times lifetime for different temperatures
    #    The vvdos_tau array has 4 dimensions (ntemp, 3, 3, nsppolplus1, nw)

    #      1. the number of temperatures
    #      2. 3x3 components of the tensor
    #      3. the spin polarization + 1 for the sum
    #      4. the number of frequencies
    #    """
    #    # nctkarr_t('vvtau_dos', "dp", "edos_nw, three, three, ntemp, nsppol, nrta")
    #    i, j = abu.s2itup(component)
    #    emesh = self.read_value("edos_mesh") * abu.Ha_eV
    #    vals = self.read_variable("vvtau_dos")
    #    vvtau_dos = vals[irta, spin, itemp, j, i, :] / (2 * abu.Ha_s)

    #    return emesh, vvtau_dos

    #def read_dos(self):
    #    """
    #    Read the density of states (in eV units)
    #    """
    #    # Total DOS, spin up and spin down component.
    #    # nctkarr_t("edos_dos", "dp", "edos_nw, nsppol_plus1")
    #    emesh = self.read_value("edos_mesh") * abu.Ha_to_eV
    #    dos = self.read_value("edos_dos") / abu.Ha_to_eV
    #    idos = self.read_value("edos_idos")

    #    #return ElectronDos(mesh, spin_dos, nelect)
    #    return emesh, dos, idos

    #def read_onsager(self, itemp):
    #    """
    #    Read the Onsager coefficients computed in the transport driver in Abinit
    #    """
    #    # nctkarr_t('L0', "dp", "edos_nw, three, three, ntemp, nsppol, nrta"), &
    #    L0 = np.moveaxis(self.read_variable("L0")[itemp,:], [0,1,2,3], [3,2,0,1])
    #    L1 = np.moveaxis(self.read_variable("L1")[itemp,:], [0,1,2,3], [3,2,0,1])
    #    L2 = np.moveaxis(self.read_variable("L2")[itemp,:], [0,1,2,3], [3,2,0,1])

    #    return L0, L1, L2

    #def read_transport(self, itemp):
    #    # nctkarr_t('sigma',   "dp", "edos_nw, three, three, ntemp, nsppol, nrta"), &
    #    sigma = np.moveaxis(self.read_variable("sigma")[itemp,:],     [0,1,2,3], [3,2,0,1])
    #    kappa = np.moveaxis(self.read_variable("kappa")[itemp,:],     [0,1,2,3], [3,2,0,1])
    #    seebeck = np.moveaxis(self.read_variable("seebeck")[itemp,:], [0,1,2,3], [3,2,0,1])
    #    pi = np.moveaxis(self.read_variable("pi")[itemp,:],           [0,1,2,3], [3,2,0,1])
    #    return sigma, kappa, seebeck, pi

    def read_mobility(self, eh, itemp, component, spin, irta=0):
        """
        Read mobility from the RTA.nc file
        The mobility is computed separately for electrons and holes.
        """
        # nctkarr_t('mobility',"dp", "three, three, edos_nw, ntemp, two, nsppol, nrta")
        i, j = abu.s2itup(component)
        wvals = self.read_variable("edos_mesh")
        #wvals = self.read_value("edos_mesh") * abu.Ha_eV
        mobility = self.read_variable("mobility")[irta, spin, eh, itemp, :, j, i]

        return wvals, mobility


class RtaRobot(Robot, RobotWithEbands):
    """
    This robot analyzes the results contained in multiple RTA.nc files.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: RtaRobot
    """

    EXT = "RTA"

    #def get_mobility_mu_dataframe(self, eh=0, component='xx', itemp=0, spin=0, **kwargs):

    @add_fig_kwargs
    def plot_mobility_kconv(self, eh=0, bte=('serta', 'mrta', 'ibte'), mode="full", component='xx', itemp=0, 
                            spin=0, fontsize=14, ax=None, **kwargs) -> Figure:
        """
        Plot the convergence of the mobility as a function of the number of k-points,
        for different transport formalisms included in the computation.

        Args:
            eh: 0 for electrons, 1 for holes.
            bte: list of transport formalism to plot (serta, mrta, ibte)
            mode: mode for the convergence plot. 'full': normal plot of the mobility
                                                 'relative': plot of the mobility relative to 
                                                             the value obtained with the densest grid
            component: Cartesian component to plot ('xx', 'xy', ...)
            itemp: temperature index.
            spin: Spin index.
            fontsize: fontsize for legends and titles
            ax: |matplotlib-Axes| or None if a new figure should be created.

        Returns: |matplotlib-Figure|
        """

        if 'ibte' in bte and not self.all_have_ibte:
            print("At least some IBTE results are missing ! Will remove ibte from the bte list")
            bte.remove('ibte')

        ax, fig, plt = get_ax_fig_plt(ax=ax)
        ax.grid(True)
        i, j = abu.s2itup(component)
        irta = 0

        res, temps = [], []
        for ncfile in self.abifiles:
            kptrlatt = ncfile.reader.read_value("kptrlatt")
            kptrlattx = kptrlatt[0, 0]
            kptrlatty = kptrlatt[1, 1]
            kptrlattz = kptrlatt[2, 2]

            mob_serta, mob_mrta, mob_ibte = -1, -1, -1
            if 'serta' in bte:
                mob_serta = ncfile.reader.read_variable("mobility_mu")[0, spin, itemp, eh, j, i]
            if 'mrta' in bte:
                mob_mrta = ncfile.reader.read_variable("mobility_mu")[1, spin, itemp, eh, j, i]
            if 'ibte' in bte:
                mob_ibte = ncfile.reader.read_variable("ibte_mob")[spin, itemp, eh, j, i]

            res.append([kptrlattx, mob_serta, mob_mrta, mob_ibte])
            temps.append(ncfile.tmesh[itemp])

        res.sort(key=lambda t: t[0])
        res = np.array(res)
        #print(res)

        size = 14
        if mode == 'relative':
            ylabel = r"Relative %s mobility (%%)" % {0: "electron", 1: "hole"}[eh]
        elif mode == 'full':
            ylabel = r"%s mobility (cm$^2$/(V$\cdot$s))" % {0: "Electron", 1: "Hole"}[eh]
        ax.set_ylabel(ylabel, size=size)

        from fractions import Fraction
        ratio1 = Fraction(kptrlatty, kptrlattx)
        ratio2 = Fraction(kptrlattz, kptrlattx)
        text1 = '' if ratio1.numerator == ratio1.denominator else \
                r'$\frac{{{0}}}{{{1}}}$'.format(ratio1.numerator, ratio1.denominator)
        text2 = '' if ratio2.numerator == ratio2.denominator else \
                r'$\frac{{{0}}}{{{1}}}$'.format(ratio2.numerator, ratio2.denominator)

        ax.set_xlabel(r'Homogeneous $N_k \times$ ' + text1 + r'$N_k \times$ ' + text2 + r'$N_k$ $\mathbf{k}$-point grid',
                      size=size)

        ax.set_title(component+" component, T = {0} K".format(ncfile.tmesh[itemp]),size=size)

        if mode == 'relative':
            for ires in [1,2,3]:
                res[:,ires] = 100*(res[:,ires]-res[-1,ires])/res[-1,ires]

        if 'serta' in bte:
            ax.plot(res[:,0], res[:,1], '-ob', label='SERTA')
        if 'mrta' in bte:
            ax.plot(res[:,0], res[:,2], '-or', label='MRTA')
        if 'ibte' in bte:
            ax.plot(res[:,0], res[:,3], '-og', label='IBTE')

        ax.set_xticks(res[:,0].astype(float))
        ax.legend(loc="best", shadow=True, fontsize=fontsize)

        return fig

    @lazy_property
    def assume_gap(self) -> bool:
        """True if we are dealing with a semiconductor. More precisely if all(sigma_erange) > 0."""
        return all(abifile.assume_gap for abifile in self.abifiles)

    @lazy_property
    def all_have_ibte(self) -> bool:
        """True if all files contain IBTE results."""
        return all(abifile.has_ibte for abifile in self.abifiles)

    def get_same_tmesh(self):
        """
        Check whether all files have the same T-mesh. Return common tmesh else raise RuntimeError.
        """
        for i in range(len(self)):
            if not np.all(self.abifiles[0].tmesh == self.abifiles[i].tmesh):
                raise RuntimeError("Found different T-mesh in RTA files.")

        return self.abifiles[0].tmesh

    #@add_fig_kwargs
    #def plot_mobility_erange_conv(self, eh=0, component='xx', itemp=0, spin=0, fontsize=8, ax=None, **kwargs):

    #@add_fig_kwargs
    #def plot_transport_tensors_mu_kconv(self, eh=0, component='xx', itemp=0, spin=0, fontsize=8, ax=None, **kwargs):

    #@add_fig_kwargs
    #def plot_transport_tensors_mu_kconv(self, eh=0, component='xx', itemp=0, spin=0, fontsize=8, ax=None, **kwargs):

    def plot_ibte_vs_rta_rho(self, component="xx", fontsize=8, **kwargs) -> Figure:
        """
        """
        nrows = 1 # xx
        ncols = len(self) # SERTA, MRTA, IBTE
        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                                sharex=True, sharey=True, squeeze=False)
        ax_list = ax_list.ravel()

        for abifile, ax in zip(self.abifiles, ax_list):
            abifile.plot_ibte_vs_rta_rho(component="xx", fontsize=fontsize, ax=ax, show=False)

        return fig

    def plot_ibte_mrta_serta_conv(self,  what="resistivity", fontsize=8, **kwargs) -> Figure:
        """
        """
        #num_plots, ncols, nrows, what_list = x2_grid(what_list)
        nrows = 1 # xx
        ncols = 3 # SERTA, MRTA, IBTE
        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                                sharex=True, sharey=True, squeeze=False)
        ax_list = ax_list.ravel()
        # don't show the last ax if numeb is odd.
        #if num_plots % ncols != 0: ax_list[-1].axis("off")

        i = j = 0
        from collections import defaultdict
        data = defaultdict(list)
        for abifile in self.abifiles:
            rta_vals = abifile.reader.read_value("resistivity")
            data["serta"].append(rta_vals[0, :, j, i])
            data["mrta"].append(rta_vals[1, :, j, i])
            ibte_vals = abifile.reader.read_value("ibte_rho")
            data["ibte"].append(ibte_vals[:, j, i])

        tmesh = self.get_same_tmesh()
        keys = ["serta", "mrta", "ibte"]
        for ix, (key, ax) in enumerate(zip(keys, ax_list)):
            ax.grid(True)
            ax.set_title(key.upper(), fontsize=fontsize)
            for ifile, ys in enumerate(data[key]):
                ax.plot(tmesh, ys, marker="o", label=self.labels[ifile])
            ax.set_xlabel("Temperature (K)")
            if ix == 0:
                ax.set_ylabel(r"Resistivity ($\mu\Omega\;cm$)")
                ax.legend(loc="best", shadow=True, fontsize=fontsize)

        return fig

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        Used in abiview.py to get a quick look at the results.
        """
        #yield self.plot_lattice_convergence(show=False)
        #if self.all_have_ibte:
        yield self.plot_ibte_mrta_serta_conv(show=False)
        yield self.plot_ibte_vs_rta_rho(show=False)

        # Determine the independent component. For the time being,
        # only consider the cubic case separately
        abifile = self.abifiles[0]
        if 'cubic' in abifile.structure.spget_summary():
            components = ['xx']
        else:
            components = ['xx','yy','zz']

        # Determine the type of carriers for which the mobility is computed
        eh_list = []
        if abifile.sigma_erange[0] > 0:
            eh_list.append(1)
        if abifile.sigma_erange[1] > 0:
            eh_list.append(0)

        for eh in eh_list:
            for comp in components:
                yield self.plot_mobility_kconv(eh=eh, component=comp, itemp=0, spin=0, show=False)

    #def get_panel(self):
    #    """
    #    Build panel with widgets to interact with the |RtaRobot| either in a notebook or in a panel app.
    #    """
    #    from abipy.panels.transportfile import TransportRobotPanel
    #    return TransportRobotPanel(self).get_panel()

    def write_notebook(self, nbpath=None) -> str:
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        args = [(l, f.filepath) for l, f in self.items()]
        nb.cells.extend([
            #nbv.new_markdown_cell("# This is a markdown cell"),
            nbv.new_code_cell("robot = abilab.RtaRobot(*%s)\nrobot.trim_paths()\nrobot" % str(args)),
            #nbv.new_code_cell("ebands_plotter = robot.get_ebands_plotter()"),
        ])

        # Mixins
        #nb.cells.extend(self.get_baserobot_code_cells())
        #nb.cells.extend(self.get_ebands_code_cells())

        return self._write_nb_nbpath(nb, nbpath)


if __name__ == "__main__":
    import sys
    robot = RtaRobot.from_files(sys.argv[1:])
    print(robot)

    #import matplotlib.pyplot as plt
    #plt.figure(0, figsize=(14,9))
    #plt.tick_params(labelsize=14)
    #ax = plt.gca()

    robot.plot_mobility_kconv(ax=None, color='k')

    #fileslist = ['conv_fine/k27x27x27/q27x27x27/Sio_DS1_TRANSPORT.nc',
    #             'conv_fine/k30x30x30/q30x30x30/Sio_DS1_TRANSPORT.nc',
    #             'conv_fine/k144x144x144/q144x144x144/Sio_DS1_TRANSPORT.nc',]

    #plot_mobility_kconv(ax, fileslist, color='k', marker='o', label=r'$N_{{q_{{x,y,z}}}}$ = $N_{{k_{{x,y,z}}}}$')

    #fileslist = ['conv_fine/k27x27x27/q54x54x54/Sio_DS1_TRANSPORT.nc',
    #             'conv_fine/k66x66x66/q132x132x132/Sio_DS1_TRANSPORT.nc',
    #             'conv_fine/k72x72x72/q144x144x144/Sio_DS1_TRANSPORT.nc']

    #plot_mobility_kconv(ax, fileslist, color='r', marker='x', label=r'$N_{{q_{{x,y,z}}}}$ = $2 N_{{k_{{x,y,z}}}}$')

    #plt.legend(loc='best',fontsize=14)
    #plt.show()



"""
Pydantic models storing the results of DFPT calculations
"""
from __future__ import annotations

from pydantic import Field
from abipy.dfpt.ddb import DdbFile
from .base_models import AbipyModel, MongoConnector #, GfsFileDesc, GfsDesc


class PhononData(AbipyModel):

    #min_phfreq_ev: float = Field(..., description="Minimum phonon frequency in eV. Includes 'negative' values")
    #max_phfreq_ev: float = Field(..., description="Maximum phonon frequency in eV.")

    #ddb_gfsd: GfsFileDesc = Field(None, description="Metadata needed to retrieve the DDB file from GridFS.")

    #phbands: PhononBands = Field(..., description="DDB string")
    #phdos: PhononDos = Field(..., description="DDB string")

    #phbands_gfsd: GfsDesc = Field(None, description="Pointer to the PhononBands object in GridFS.")
    #phdos_gfsd: GfsDesc = Field(None, description="Pointer to the PhononDos object in GridFS.")

    @classmethod
    def from_ddb(cls, ddb: DdbFile, mng_connector: MongoConnector) -> PhononData:
        """
        Build the model from a DdbFile `ddb`.
        """
        data = dict(
            #ddb_gfsd=mng_connector.gfs_put_filepath(ddb.filepath)
        )

        # TODO: Add Msonable protocol to phbands and phdos.
        # Should we store PHBST, PHDOS and anaddb.nc or just MSonable objects ?

        #with ddb.anaget_phbst_and_phdos_files() as g:
        #    phbst_file, phdos_file = g[0], g[1]
        #    phbands = phbst_file.phbands
        #    phdos = phdos_file.phdos
        #    data.update(dict(
        #        phbands=phbands,
        #        phdos=phdos,
        #
        #    ))
        #    self.phbands_gfsd = mng_connector.gfs_put_mson(phbst_file.phbands)
        #    self.phdos_gfsd = mng_connector.gfs_put_mson(phdos_file.phdos)

        return cls(**data)



# coding: utf-8
"""
Task and Work subclasses related to structure relaxations.
"""
from __future__ import annotations

from abipy.core.structure import Structure
from abipy.dynamics.hist import HistRobot
from abipy.abio.inputs import AbinitInput
from abipy.flowtk.tasks import RelaxTask
from abipy.flowtk.works import Work


class RelaxTaskWithTargetDilatmx(RelaxTask):
    """
    Relaxation Task that iteratively reduces the value of
    dilatmx until a target value is reached.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: RelaxTaskWithTargetDilatmx
    """

    @classmethod
    def from_scf_input(cls, scf_input: AbinitInput,
                       target_dilatmx=1.001, workdir=None, manager=None) -> RelaxTaskWithTargetDilatmx
        """
        Build an instance from an Abinit input representing a GS calculation.
        """
        relax_input = scf_input.new_with_vars(
            optcell=2,
            ionmov=22,
            ecutsm=0.5,
            dilatmx=1.1,
        )

        task = cls(relax_input, workdir=workdir, manager=manager)
        task.target_dilatmx = target_dilatmx

        return task

    def on_ok(self):
        """
        This method is called once the `Task` has reached status S_OK.
        """
        actual_dilatmx = self.input["dilatmx"]
        if self.target_dilatmx < actual_dilatmx:
            self.reduce_dilatmx(self.target_dilatmx)
            self.history.info('Converging dilatmx. Value reduced from {} to {}.'
                               .format(actual_dilatmx, self.input['dilatmx']))
            self.restart()
            self.finalized = False
            return dict(returncode=0, message="Restarting task with smaller dilatmx")
        else:
            self.history.info(f"Reached target dilatmx: {self.target_dilatmx}. Finalized set to True")
            self.finalized = True
            return dict(returncode=0, message="Restarting task with smaller dilatmx")


class RelaxTaskWithM3gnet(RelaxTask):
    """
    Relaxation Task that performs an initial relaxation with m3gnet
    followed by an ab-initio relaxation with Abinit.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: RelaxTaskWithM3gnet
    """

    @classmethod
    def from_scf_input(cls, scf_input: AbinitInput) -> RelaxTaskWithM3gnet
        """
        Build an instance from an Abinit input representing a GS calculation.
        """
        # Run m3gnet.
        relaxed_structure = scf_input.structure.relax(calculator="m3gnet", relax_cell=True)
        relaxed_structure.__class__ = Structure

        #relaxed_structure = relaxed_structure.refine()
        relaxed_structure = relaxed_structure.abi_sanitize()

        relax_input = scf_input.new_with_vars(
            ionmov=22,
            optcell=2,
            ecutsm=0.5,
            dilatmx=1.1,
            #toldff=1e-6,
            #tolmxf=1e-5,
            #ntime=100,
        )

        relax_input.set_structure(relaxed_structure)

        return cls(relax_input)


class AbiVsM3gnetWork(Work):
    """
    This work is mainly used for benchmarking/testing purposes.
    It contains two tasks performing two structural relaxations.
    The first task performs a full ab-initio relaxation starting from the input structure,
    The second one performs an initial relaxation using m3gnet before calling Abinit.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: AbiVsM3gnetWork
    """

    @classmethod
    def from_relax_input(cls, relax_input: AbinitInput) -> AbiVsM3gnetWork:
        """Build the work from an input for structura relaxations."""
        work = cls()
        work.abi_relax_task = work.register_relax_task(relax_input)
        work.m3g_relax_task = m3g_task = RelaxTaskWithM3gnet(relax_input)
        work.register_task(m3g_task)
        return work

    def on_all_ok(self):
        """
        This method is called once the `Work` is completed i.e. when all tasks have reached status S_OK.
        """
        filepaths = [work.abi_relax_task.hist_path, work.m3g_relax_task.hist_path]
        with HistRobot.from_files(filepaths) as robot:
            print(robot)
            df = robot.get_dataframe()
            df.to_excel(self.outdir.path_in("abivsm3gnet.xlsx"))

            with robot.get_pyscript(self.outdir.path_in("hist_robot.py")) as script:
                script.add_text("""
df = robot.get_dataframe()
for what in robot.what_list:
    robot.gridplot(what=what, tight_layout=True)
    #robot.combiplot(what=what, tight_layout=True)
""")

        return super().on_all_ok()



#!/usr/bin/env python
"""Compare GS calculations with NC pseudos performed with useylm in [0, 1]."""
import sys
import abipy.abilab as abilab
import abipy.flowtk as flowtk
import abipy.data as abidata

from itertools import product
from abipy.benchmarks import bench_main, BenchmarkFlow


def make_input():
    """Build a template input file for GS calculations with k-point parallelism """
    pseudos = abidata.pseudos("14si.pspnc", "8o.pspnc")

    structure = abidata.structure_from_ucell("SiO2-alpha")
    #structure.make_supercell([3, 3, 3])

    inp = abilab.AbinitInput(structure, pseudos)
    inp.set_kmesh(ngkpt=[4, 4, 4], shiftk=[0, 0, 0])

    # Global variables
    ecut = 24
    inp.set_vars(
        ecut=ecut,
        nsppol=1,
        nband=28,
        paral_kgb=0,
        istwfk="*1",
        timopt=-1,
        chksymbreak=0,
        chkprim=0,
        maxnsym=2400,
        prtwf=0,
        prtden=0,
        tolvrs=1e-8,
    )

    return inp


def build_flow(options):
    inp = make_input()

    mpi_list = options.mpi_list
    if mpi_list is None:
        nkpt = len(inp.abiget_ibz().points)
        nks = nkpt * inp["nsppol"]
        mpi_list = [p for p in range(1, nks + 1) if nks % p == 0]
    if options.verbose: print("Using mpi_list:", mpi_list)

    flow = BenchmarkFlow(workdir=options.get_workdir(__file__), remove=options.remove)

    for useylm in [0, 1]:
        work = flowtk.Work()
        for mpi_procs, omp_threads in product(mpi_list, options.omp_list):
            if not options.accept_mpi_omp(mpi_procs, omp_threads): continue
            manager = options.manager.new_with_fixed_mpi_omp(mpi_procs, omp_threads)
            work.register_scf_task(inp.new_with_vars(useylm=useylm), manager=manager)
        flow.register_work(work)

    return flow.allocate()


@bench_main
def main(options):
    if options.info:
        # print doc string and exit.
        print(__doc__)
        return

    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
"""Optical spectra with Optic."""
from __future__ import print_function, division, unicode_literals, absolute_import

import sys
import os
import abipy.data as abidata
import abipy.abilab as abilab
import abipy.flowtk as flowtk


def build_flow(options, paral_kgb=0):
    """
    Build flow for the calculation of optical properties with optic + band structure
    along high-symmetry k-path. DDK are computed with 3 k-meshes of increasing density
    to monitor the convergece of the spectra.
    """
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    workdir = options.workdir
    if not options.workdir:
        workdir = os.path.basename(__file__).replace(".py", "").replace("run_", "flow_")

    multi = abilab.MultiDataset(structure=abidata.structure_from_ucell("GaAs"),
                                pseudos=abidata.pseudos("31ga.pspnc", "33as.pspnc"), ndtset=2)

    # Usa same shifts in all tasks.
    shiftk= [[0.5, 0.5, 0.5],
             [0.5, 0.0, 0.0],
             [0.0, 0.5, 0.0],
             [0.0, 0.0, 0.5]]

    # Global variables.
    multi.set_vars(ecut=2, paral_kgb=paral_kgb)

    # Dataset 1 (GS run)
    multi[0].set_vars(tolvrs=1e-8, nband=4)
    multi[0].set_kmesh(ngkpt=[4, 4, 4], shiftk=shiftk)

    # NSCF run on k-path with large number of bands
    multi[1].set_vars(iscf=-2, nband=20, tolwfr=1.e-9)
    multi[1].set_kpath(ndivsm=10)

    # Initialize the flow.
    flow = flowtk.Flow(workdir, manager=options.manager, remove=options.remove)

    # GS to get the density + NSCF along the path.
    scf_inp, nscf_inp = multi.split_datasets()
    bands_work = flowtk.BandStructureWork(scf_inp, nscf_inp)
    flow.register_work(bands_work)

    # Build OpticInput used to compute optical properties.
    optic_input = abilab.OpticInput(
        broadening=0.002,          # Value of the smearing factor, in Hartree
        domega=0.0003,             # Frequency mesh.
        maxomega=0.3,
        scissor=0.000,             # Scissor shift if needed, in Hartree
        tolerance=0.002,           # Tolerance on closeness of singularities (in Hartree)
        num_lin_comp=2,            # Number of components of linear optic tensor to be computed
        lin_comp=(11, 33),         # Linear coefficients to be computed (x=1, y=2, z=3)
        num_nonlin_comp=2,         # Number of components of nonlinear optic tensor to be computed
        nonlin_comp=(123, 222),    # Non-linear coefficients to be computed
        num_linel_comp=1,          # Number of components of LEO tensor to be computed
        linel_comp=(123),          # Non-linear coefficients to be computed
    )

    # ddk_nband is fixed here, in principle it depends on nelect and the frequency range in chi(w).
    ddk_nband = 20

    # Perform converge study wrt ngkpt (shiftk is constant).
    ngkpt_convergence = [[4, 4, 4], [8, 8, 8], [12, 12, 12]]

    for ddk_ngkpt in ngkpt_convergence:
        # Build work for NSCF from DEN produced by the first GS task + 3 DDKs.
        # All tasks use more bands and a denser k-mesh defined by ddk_ngkpt.
        ddks_work = flowtk.NscfDdksWork.from_scf_task(bands_work[0], ddk_ngkpt, shiftk, ddk_nband)
        flow.register_work(ddks_work)

        # Build optic task to compute chi with this value of ddk_ngkpt.
        optic_task = flowtk.OpticTask(optic_input, nscf_node=ddks_work.task_with_ks_energies,
                                      ddk_nodes=ddks_work.ddk_tasks, use_ddknc=False)
        ddks_work.register_task(optic_task)

    return flow


@flowtk.flow_main
def main(options):
    return build_flow(options)


if __name__ == "__main__":
    retcode = main()
    if retcode != 0: sys.exit(retcode)

    rename_table = [
        #  src, dest
        #"_runflow/w0/t1/outdata/out_GSR.nc",
        ("_runflow/w1/t4/outdata/out_OPTIC.nc", "gaas_444_OPTIC.nc"),
        ("_runflow/w2/t4/outdata/out_OPTIC.nc", "gaas_888_OPTIC.nc"),
        ("_runflow/w3/t4/outdata/out_OPTIC.nc", "gaas_121212_OPTIC.nc"),
        ("_runflow/w1/t1/outdata/out_DDK.nc", "gaas_444_dir1_DDK.nc"),
        ("_runflow/w1/t2/outdata/out_DDK.nc", "gaas_444_dir2_DDK.nc"),
        ("_runflow/w1/t3/outdata/out_DDK.nc", "gaas_444_dir3_DDK.nc"),
    ]
    import shutil
    for old, new in rename_table:
        shutil.copyfile(old, new)

    #shutil.rmtree("_runflow")
    sys.exit(0)



"""
This module provides objects related to site symmetries
"""
from __future__ import annotations

import numpy as np
import pandas as pd
import sympy as sp

from collections import OrderedDict
from monty.termcolor import cprint
from abipy.core.mixins import Has_Structure
from abipy.core.structure import Structure


class SiteSymmetries(Has_Structure):

    def __init__(self, structure: Structure):
        """
        Args:
            structure: |Structure| object.
        """
        # Get spacegroup from spglib if not available from Abinit.
        if not structure.has_abi_spacegroup:
            structure.spgset_abi_spacegroup(has_timerev=True, overwrite=False)

        self._structure = structure

        abispg = structure.abi_spacegroup
        nsym = len(abispg.symrel)
        indsym = self.structure.indsym

        #self.eq_atoms = structure.spget_equivalent_atoms()

        # Precompute sympy objects used to solve linear system of equations.
        self.sp_symrel, self.sp_symrec = [], []
        self.sp_tnons, self.sp_inv_symrel = [], []
        from abipy.core.symmetries import mati3inv

        for symr, symc, tau in zip(abispg.symrel, abispg.symrec, abispg.tnons):
            self.sp_symrel.append(sp.Matrix((symr)))
            inv_symr = mati3inv(symr, trans=False)
            self.sp_inv_symrel.append(sp.Matrix((inv_symr)))
            self.sp_symrec.append(sp.Matrix((symc)))
            # FIXME: Should convert to rational numbers
            # Permissible translations are unit cell translations or fractions thereof
            # that are consistent with the rotational symmetry (e.g. 1/2, 1/3, 1/4, and 1/6), plus combinations.
            tau = np.around(tau, decimals=5)
            self.sp_tnons.append(sp.Matrix(tau))

        #from abipy.core.symmetries import indsym_from_symrel
        #other_indsym = indsym_from_symrel(abispg.symrel, abispg.tnons, self.structure, tolsym=1e-8)
        #assert np.all(self.structure.indsym == other_indsym)

        # Compute symmetry operations in Cartesian coordinates (numpy arrays)
        a = self.structure.lattice.matrix.T
        self.symcart = np.matmul(a, np.matmul(abispg.symrel, np.linalg.inv(a)))

        import spglib
        self.sitesym_labels = []
        for iatom, site in enumerate(self.structure):
            rotations = [abispg.symrel[isym] for isym in range(nsym) if
                         indsym[iatom, isym, 3] == iatom and abispg.symafm[isym] == 1]
            # Passing a 0-length rotations list to spglib can segfault.
            herm_symbol, ptg_num = "1", 1
            if len(rotations) != 0:
                herm_symbol, ptg_num, trans_mat = spglib.get_pointgroup(rotations)

            self.sitesym_labels.append("%s (#%d,nsym:%d)" % (herm_symbol.strip(), ptg_num, len(rotations)))

        #for irred_isite in self.irred_isites:
        #    for isite_eq, rm1, tau, l0 in self.eq_sites[irred_isite]:
        #        # isite_eq = rm1(irred_site - tau) + l0

    @property
    def structure(self) -> Structure:
        """|Structure| object."""
        return self._structure

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self, verbose=0) -> str:
        """String representation with verbosity level verbose."""
        lines = []; app = lines.append
        app(self.structure.to_string(verbose=verbose, title="Structure"))

        return "\n".join(lines)

    def get_wyckoff_dataframe(self, view="all", select_symbols=None, decimals=5, verbose=0) -> pd.DataFrame:
        """
        Find Wyckoff positions.

        Args:
            view:
            select_symbols:
            decimals: Number of decimal places to round to.
                If decimals is negative, it specifies the number of positions to the left of the decimal point.
            verbose: Verbosity level.

        Return |pandas-DataFrame| with cartesian tensor components as columns and (inequivalent) sites along the rows.
        """
        # Select atoms.
        aview = self._get_atomview(view, select_symbols, verbose=verbose)

        sitesym_labels = self.structure.spget_site_symmetries()

        frac_symbols = sp.symbols("xfrac, yfrac, zfrac")
        vector = sp.Matrix((frac_symbols))

        indsym = self.structure.indsym
        abispg = self.structure.abi_spacegroup
        rows = []
        for (iatom, wlabel) in zip(aview.iatom_list, aview.wyck_labels):
            site = self.structure[iatom]
            system = []
            for isym, (rm1, tau) in enumerate(zip(self.sp_inv_symrel, self.sp_tnons)):
                if indsym[iatom, isym, 3] != iatom: continue
                l0 = indsym[iatom, isym, :3]
                l0 = sp.Matrix(l0)
                m = rm1 * (vector - tau) - (l0 + vector)
                if verbose:
                    print(92 * "=")
                    print("System of linear equations for iatom %d, %s" % (iatom, repr(site)))
                    sp.pprint(m)
                    print("Rotation:")
                    sp.pprint(rm1)

                system.append(m)

            # Solve system of linear equations.
            solutions = sp.solve(system, dict=True)
            #print(solutions)
            if verbose and not solutions:
                cprint("No solution for iatom %d" % iatom, "yellow")

            if solutions:
                d = OrderedDict()
                d["element"] = site.specie.symbol
                d["site_index"] = iatom
                d["cart_coords"] = np.round(site.coords, decimals=decimals)
                d["frac_coords"] = np.round(site.frac_coords, decimals=decimals)
                d["wyckoff"] = wlabel
                d["site_symmetry"] = sitesym_labels[iatom]
                for s in frac_symbols:
                    d[str(s)] = str(s)
                if len(solutions) > 1:
                    cprint("Found multiple solutions for iatom %d" % iatom, "red")
                d.update({str(k): str(v) for k, v in solutions[0].items()})
                rows.append(d)

        df = pd.DataFrame(rows, index=None, columns=list(rows[0].keys()) if rows else None)
        return df

    def get_tensor_rank2_dataframe(self, view="all", select_symbols=None, decimals=5, verbose=0) -> pd.DataFrame:
        """
        Use site symmetries to detect indipendent elements of rank 2 tensor.

        Args:
            view:
            select_symbols:
            decimals: Number of decimal places to round to.
                If decimals is negative, it specifies the number of positions to the left of the decimal point.
            verbose: Verbosity level

        Return |pandas-DataFrame| with cartesian tensor components as columns and (inequivalent) sites along the rows.
        """
        # Select atoms.
        aview = self._get_atomview(view, select_symbols, verbose=verbose)

        sitesym_labels = self.structure.spget_site_symmetries()

        # Symmetric tensor in reduced coordinates (direct lattice)
        # Operations in reduced coords are given by integer matrices and this facilitates the solution
        # of the system of equations with sympy.
        Txx, Tyy, Tzz, Txy, Txz, Tyz = symbols = sp.symbols('Txx, Tyy, Tzz, Txy, Txz, Tyz')
        tensor = sp.Matrix(([Txx, Txy, Txz], [Txy, Tyy, Tyz], [Txz, Tyz, Tzz]))

        indsym = self.structure.indsym
        rows = []
        for (iatom, wlabel) in zip(aview.iatom_list, aview.wyck_labels):
            site = self.structure[iatom]
            system = []
            for isym, rotf in enumerate(self.sp_symrel):
                if indsym[iatom, isym, 3] != iatom: continue
                m = rotf * tensor * rotf.T - tensor
                if verbose:
                    print(92 * "=")
                    print("System of linear equations for iatom %d, %s" % (iatom, repr(site)))
                    sp.pprint(m)
                    print("Rotation:")
                    sp.pprint(rotf)
                    print(92 * "=")

                system.append(m)

            # Solve system of linear equations. Print only sites for which we have contraints.
            solutions = sp.solve(system, dict=True)
            #print(solutions)
            if verbose and not solutions:
                cprint("No solution for iatom %d" % iatom, "yellow")

            if solutions:
                d = OrderedDict()
                d["element"] = site.specie.symbol
                d["site_index"] = iatom
                d["frac_coords"] = np.round(site.frac_coords, decimals=decimals)
                d["cart_coords"] = np.round(site.coords, decimals=decimals)
                d["wyckoff"] = wlabel
                d["site_symmetry"] = sitesym_labels[iatom]
                for s in symbols:
                    d[str(s)] = str(s)
                if len(solutions) > 1:
                    cprint("Found multiple solutions for iatom %d" % iatom, "red")
                d.update({str(k): str(v) for k, v in solutions[0].items()})
                rows.append(d)

        import pandas as pd
        df = pd.DataFrame(rows, index=None, columns=list(rows[0].keys()) if rows else None)
        return df

    def check_site_symmetries(self, tcart, verbose=0) -> float:
        """
        Test whether a set of tensors associated to the crystalline sites are compatible
        with the space group symmetries.

        Args:
            tcart: (natom, 3, 3) array
            verbose: Verbosity level.

        Return: max_err
        """
        natom = len(self.structure)
        indsym = self.structure.indsym
        nsym = len(self.symcart)
        tcart = np.reshape(tcart, (natom, 3, 3))

        max_err = 0.0
        for iatom in range(natom):
            ref_mat = tcart[iatom]
            sym_mat = np.zeros_like(ref_mat)
            count = 0
            for isym, scart in enumerate(self.symcart):
                if indsym[iatom, isym, 3] != iatom: continue
                count += 1
                sym_mat += np.matmul(scart, np.matmul(ref_mat, scart.T))
                #sym_mat += np.matmul(scart.T, np.matmul(ref_mat, scart))

            if (nsym // count) * count != nsym: max_err = 1e+23
            sym_mat /= count
            diff_mat = sym_mat - ref_mat
            max_err = max(max_err, np.abs(diff_mat).sum())
            if count != 1 and verbose:
                print("For iatom", iatom, "on-site symmetries with count:", count)
                print("ref_mat:\n", ref_mat, "\nsym_mat:\n", sym_mat, "\ndiff_mat:\n", diff_mat)

        for iatom in range(natom):
            ref_mat = tcart[iatom]
            for isym, scart in enumerate(self.symcart):
                jatom = indsym[iatom, isym, 3]
                if jatom == iatom: continue
                #sym_mat = np.matmul(scart, np.matmul(ref_mat, scart.T))
                sym_mat = np.matmul(scart.T, np.matmul(ref_mat, scart))
                diff_mat = sym_mat - tcart[jatom]
                max_err = max(max_err, np.abs(diff_mat).sum())
                if verbose:
                    print("For iatom", iatom, "ref_mat, sym_mat, diff_mat")
                    print("ref_mat:\n", tcart[jatom], "\nsym_mat:\n", sym_mat, "\ndiff_mat:\n", diff_mat)

        if verbose: print("Max symmetrization error:", max_err)

        return max_err



""""Utilities for AbiPy panels."""
from __future__ import annotations

import panel as pn
import panel.widgets as pnw

def btn_open_link(url: str, new_tab=True, **btn_kwargs) -> pn.Button:
    """
    Return button to open link in a new tab.
    """
    name = btn_kwargs.get("name", "Open Link")
    btn = pnw.Button(name=name, **btn_kwargs)
    btn.js_on_click(code=f"window.open('{url}')" if new_tab else f"window.location.href='{url}'")

    return btn



from __future__ import annotations

import os
import socket

from contextlib import closing


def find_free_port(address: str = "localhost") -> int:
    """Retur free port"""
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind((address, 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def port_is_open(port: int, address: str = "localhost") -> bool:
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        if sock.connect_ex((address, port)) == 0: return True
        return False


def pid_exists(pid: int) -> bool:
    """
    Check whether pid exists in the current process table. UNIX only.

    https://stackoverflow.com/questions/568271/how-to-check-if-there-exists-a-process-with-a-given-pid-in-python/6940314#6940314
    """
    # import os
    import errno
    if pid < 0:
        return False
    if pid == 0:
        # According to "man 2 kill" PID 0 refers to every process
        # in the process group of the calling process.
        # On certain systems 0 is a valid PID but we have no way
        # to know that in a portable fashion.
        raise ValueError('invalid PID 0')
    try:
        os.kill(pid, 0)
    except OSError as err:
        if err.errno == errno.ESRCH:
            # ESRCH == No such process
            return False
        elif err.errno == errno.EPERM:
            # EPERM clearly means there's a process to deny access to
            return True
        else:
            # According to "man 2 kill" possible error values are
            # (EINVAL, EPERM, ESRCH)
            raise
    else:
        return True
